# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/RubanSeven/Text-Image-Augmentation-python/blob/master/augment.py
"""

import numpy as np
from .warp_mls import WarpMLS


def tia_distort(src, segment=4):
    img_h, img_w = src.shape[:2]

    cut = img_w // segment
    thresh = cut // 3

    src_pts = list()
    dst_pts = list()

    src_pts.append([0, 0])
    src_pts.append([img_w, 0])
    src_pts.append([img_w, img_h])
    src_pts.append([0, img_h])

    dst_pts.append([np.random.randint(thresh), np.random.randint(thresh)])
    dst_pts.append(
        [img_w - np.random.randint(thresh), np.random.randint(thresh)])
    dst_pts.append(
        [img_w - np.random.randint(thresh), img_h - np.random.randint(thresh)])
    dst_pts.append(
        [np.random.randint(thresh), img_h - np.random.randint(thresh)])

    half_thresh = thresh * 0.5

    for cut_idx in np.arange(1, segment, 1):
        src_pts.append([cut * cut_idx, 0])
        src_pts.append([cut * cut_idx, img_h])
        dst_pts.append([
            cut * cut_idx + np.random.randint(thresh) - half_thresh,
            np.random.randint(thresh) - half_thresh
        ])
        dst_pts.append([
            cut * cut_idx + np.random.randint(thresh) - half_thresh,
            img_h + np.random.randint(thresh) - half_thresh
        ])

    trans = WarpMLS(src, src_pts, dst_pts, img_w, img_h)
    dst = trans.generate()

    return dst


def tia_stretch(src, segment=4):
    img_h, img_w = src.shape[:2]

    cut = img_w // segment
    thresh = cut * 4 // 5

    src_pts = list()
    dst_pts = list()

    src_pts.append([0, 0])
    src_pts.append([img_w, 0])
    src_pts.append([img_w, img_h])
    src_pts.append([0, img_h])

    dst_pts.append([0, 0])
    dst_pts.append([img_w, 0])
    dst_pts.append([img_w, img_h])
    dst_pts.append([0, img_h])

    half_thresh = thresh * 0.5

    for cut_idx in np.arange(1, segment, 1):
        move = np.random.randint(thresh) - half_thresh
        src_pts.append([cut * cut_idx, 0])
        src_pts.append([cut * cut_idx, img_h])
        dst_pts.append([cut * cut_idx + move, 0])
        dst_pts.append([cut * cut_idx + move, img_h])

    trans = WarpMLS(src, src_pts, dst_pts, img_w, img_h)
    dst = trans.generate()

    return dst


def tia_perspective(src):
    img_h, img_w = src.shape[:2]

    thresh = img_h // 2

    src_pts = list()
    dst_pts = list()

    src_pts.append([0, 0])
    src_pts.append([img_w, 0])
    src_pts.append([img_w, img_h])
    src_pts.append([0, img_h])

    dst_pts.append([0, np.random.randint(thresh)])
    dst_pts.append([img_w, np.random.randint(thresh)])
    dst_pts.append([img_w, img_h - np.random.randint(thresh)])
    dst_pts.append([0, img_h - np.random.randint(thresh)])

    trans = WarpMLS(src, src_pts, dst_pts, img_w, img_h)
    dst = trans.generate()

    return dst


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/02 20:16

"""
xml等网页结构方面的处理
"""

# from pyxllib.prog.pupil import check_install_package

# 一个xpath解析库
# check_install_package('xpath_parser', 'xpath-parser')

import collections
from collections import Counter, defaultdict
import re
import os

import requests
import pandas as pd
import bs4
from bs4 import BeautifulSoup
from humanfriendly import format_size
# from xpath_parser import XpathExpression

from pyxllib.prog.newbie import round_int
from pyxllib.prog.pupil import dprint, run_once, inject_members
from pyxllib.prog.specialist import browser
from pyxllib.algo.pupil import SearchBase
from pyxllib.text.newbie import xldictstr
from pyxllib.text.pupil import shorten, ensure_gbk, BookContents, strwidth, grp_chinese_char
from pyxllib.file.specialist import File, Dir, get_etag
from pyxllib.text.jinjalib import get_jinja_template


class XlBs4Tag(bs4.element.Tag):

    @property
    def tag_name(self):
        """输入一个bs4的Tag或NavigableString，
        返回tag.name或者'NavigableString'
        """
        if self.name:
            return self.name
        elif isinstance(self, bs4.element.NavigableString):
            return 'NavigableString'
        else:
            dprint(self)  # 获取结点t名称失败
            return None

    def subtag_names(self):
        """ 列出结点的所有直接子结点（花括号后面跟的数字是连续出现次数）
        例如body的： p{137}，tbl，p{94}，tbl，p{1640}，sectPr
        """

        def counter(m):
            s1 = m.group(1)
            n = (m.end(0) - m.start(0)) // len(s1)
            s = s1[:-1] + '{' + str(n) + '}'
            if m.string[m.end(0) - 1] == '，':
                s += '，'
            return s

        if self.name and self.contents:
            s = '，'.join([x.tag_name for x in self.contents]) + '，'
            s = re.sub(r'([^，]+，)(\1)+', counter, s)
        else:
            s = ''
        if s and s[-1] == '，':
            s = s[:-1]
        return s

    def treestruct_raw(self, **kwargs):
        """ 查看树形结构的raw版本
        各参数含义详见dfs_base
        """
        # 1 先用dfs获得基本结果
        sb = SearchBase(self)
        s = sb.fmt_nodes(**kwargs)
        return s

    def treestruct_brief(self, linenum=True, prefix='- ', **kwargs):
        """ 查看树形结构的简洁版
        """

        class Search(SearchBase):
            def fmt_node(self, node, depth, *, prefix=prefix, show_node_type=False):
                if isinstance(node, bs4.element.ProcessingInstruction):
                    s = 'ProcessingInstruction，' + str(node)
                elif isinstance(node, bs4.element.Tag):
                    s = node.name + '，' + xldictstr(node.attrs, item_delimit='，')
                elif isinstance(node, bs4.element.NavigableString):
                    s = shorten(str(node), 200)
                    if not s.strip():
                        s = '<??>'
                else:
                    s = '遇到特殊类型，' + str(node)
                return (prefix * depth) + s

        search = Search(self)
        res = search.fmt_nodes(linenum=linenum, **kwargs)
        return res

    def treestruct_stat(self):
        """生成一个两个二维表的统计数据
            ls1, ls2 = treestruct_stat()
                ls1： 结点规律表
                ls2： 属性规律表
        count_tagname、check_tag的功能基本都可以被这个函数代替
        """

        def text(t):
            """ 考虑到结果一般都是存储到excel，所以会把无法存成gbk的字符串删掉
            另外控制了每个元素的长度上限
            """
            s = ensure_gbk(t)
            s = s[:100]
            return s

        def depth(t):
            """结点t的深度"""
            return len(tuple(t.parents))

        t = self.contents[0]
        # ls1 = [['element序号', '层级', '结构', '父结点', '当前结点', '属性值/字符串值', '直接子结点结构']]
        # ls2 = [['序号', 'element序号', '当前结点', '属性名', '属性值']]  #
        ls1 = []  # 这个重点是分析结点规律
        ls2 = []  # 这个重点是分析属性规律
        i = 1
        while t:
            # 1 结点规律表
            d = depth(t)
            line = [i, d, '_' * d + str(d), t.parent.tag_name, t.tag_name,
                    text(xldictstr(t.attrs) if t.name else t),  # 结点存属性，字符串存值
                    t.subtag_names()]
            ls1.append(line)
            # 2 属性规律表
            if t.name:
                k = len(ls2)
                for attr, value in t.attrs.items():
                    ls2.append([k, i, t.tag_name, attr, value])
                    k += 1
            # 下个结点
            t = t.next_element
            i += 1
        df1 = pd.DataFrame.from_records(ls1,
                                        columns=['element序号', '层级', '结构', '父结点', '当前结点', '属性值/字符串值',
                                                 '直接子结点结构'])
        df2 = pd.DataFrame.from_records(ls2, columns=['序号', 'element序号', '当前结点', '属性名', '属性值'])
        return df1, df2

    def count_tagname(self):
        """统计每个标签出现的次数：
             1                    w:rpr  650
             2                 w:rfonts  650
             3                   w:szcs  618
             4                      w:r  565
             5                     None  532
             6                      w:t  531
        """
        ct = collections.Counter()

        def inner(node):
            try:
                ct[node.name] += 1
                for t in node.children:
                    inner(t)
            except AttributeError:
                pass

        inner(self)
        return ct.most_common()

    def check_tag(self, tagname=None):
        """ 统计每个标签在不同层级出现的次数：

        :param tagname:
            None：统计全文出现的各种标签在不同层级出现次数
            't'等值： tagname参数允许只检查特殊标签情况，此时会将所有tagname设为第0级

        TODO 检查一个标签内部是否有同名标签？
        """
        d = defaultdict()

        def add(name, depth):
            if name not in d:
                d[name] = defaultdict(int)
            d[name][depth] += 1

        def inner(node, depth):
            if isinstance(node, bs4.element.ProcessingInstruction):
                add('ProcessingInstruction', depth)
            elif isinstance(node, bs4.element.Tag):
                if node.name == tagname and depth:
                    dprint(node, depth)  # tagname里有同名子标签
                add(node.name, depth)
                for t in node.children:
                    inner(t, depth + 1)
            elif isinstance(node, bs4.element.NavigableString):
                add('NavigableString', depth)
            else:
                add('其他特殊结点', depth)

        # 1 统计结点在每一层出现的次数
        if tagname:
            for t in self.find_all(tagname):
                inner(t, 0)
        else:
            inner(self, 0)

        # 2 总出现次数和？

        return d

    def check_namespace(self):
        """检查名称空间问题，会同时检查标签名和属性名：
            1  cNvPr  pic:cNvPr(579)，wps:cNvPr(52)，wpg:cNvPr(15)
            2   spPr                   pic:spPr(579)，wps:spPr(52)
        """
        # 1 获得所有名称
        #    因为是采用node的原始xml文本，所以能保证会取得带有名称空间的文本内容
        ct0 = Counter(re.findall(r'<([a-zA-Z:]+)', str(self)))
        ct = defaultdict(str)
        s = set()
        for key, value in ct0.items():
            k = re.sub(r'.*:', '', key)
            if k in ct:
                s.add(k)
                ct[k] += f'，{key}({value})'
            else:
                ct[k] = f'{key}({value})'

        # 2 对有重复和无重复的元素划分存储
        ls1 = []  # 有重复的存储到ls1
        ls2 = []  # 没有重复的正常结果存储到ls2，可以不显示
        for k, v in ct.items():
            if k in s:
                ls1.append([k, v])
            else:
                ls2.append([k, v])

        # 3 显示有重复的情况
        # browser(ls1, filename='检查名称空间问题')
        return ls1

    def get_catalogue(self, *args, size=False, start_level=-1, **kwargs):
        """ 找到所有的h生成文本版的目录

        :param bool|int size: 布尔或者乘因子，表示是否展示文本，以及乘以倍率，比如双语阅读时，size可以缩放一半

        *args, **kwargs 参考 BookContents.format_str

        注意这里算法跟css样式不太一样，避免这里能写代码，能做更细腻的操作
        """
        bc = BookContents()
        for h in self.find_all(re.compile(r'h\d')):
            if size:
                part_size = h.section_text_size(size, fmt=True)
                bc.add(int(h.name[1]), h.get_text().replace('\n', ' ').strip(), part_size)
            else:
                bc.add(int(h.name[1]), h.get_text().replace('\n', ' ').strip())

        if 'page' not in kwargs:
            kwargs['page'] = size

        if bc.contents:
            return bc.format_str(*args, start_level=start_level, **kwargs)
        else:
            return ''

    def section_text_size(self, factor=1, fmt=False):
        """ 计算某节标题下的正文内容长度 """
        if not re.match(r'h\d+$', self.name):
            raise TypeError

        # 这应该是相对比较简便的计算每一节内容多长的算法~~
        part_size = 0
        for x in self.next_siblings:
            if x.name == self.name:
                break
            else:
                text = str(x) if isinstance(x, bs4.element.NavigableString) else x.get_text()
                part_size += strwidth(text)
        part_size = round_int(part_size * factor)

        if fmt:
            return format_size(part_size).replace(' ', '').replace('bytes', 'B')
        else:
            return part_size

    def head_add_size(self, factor=1):
        """ 标题增加每节内容大小标记

        :param factor: 乘因子，默认是1。但双语阅读等情况，内容会多拷贝一份，此时可以乘以0.5，显示正常原文的大小。
        """
        for h in self.find_all(re.compile(r'h\d')):
            part_size = h.section_text_size(factor, fmt=True)
            navi_str = list(h.strings)[-1].rstrip()
            navi_str.replace_with(str(navi_str) + '，' + part_size)

    def head_add_number(self, start_level=-1, jump=True):
        """ 标题增加每节编号
        """
        bc = BookContents()
        heads = list(self.find_all(re.compile(r'h\d')))
        for h in heads:
            bc.add(int(h.name[1]), h.get_text().replace('\n', ' '))

        if not bc.contents:
            return

        nums = bc.format_numbers(start_level=start_level, jump=jump)
        for i, h in enumerate(heads):
            navi_strs = list(h.strings)
            if navi_strs:
                navi_str = navi_strs[0]
                if nums[i]:
                    navi_str.replace_with(nums[i] + ' ' + str(navi_str))
            else:
                h.string = nums[i]

    def xltext(self):
        """ 自己特用的文本化方法

        有些空格会丢掉，要用这句转回来

        210924周五20:23，但后续实验又遭到了质疑，目前这功能虽然留着，但不建议使用
        """
        # return self.prettify(formatter=lambda s: s.replace(u'\xa0', '&nbsp;'))
        # \xa0好像是些特殊字符，删掉就行。。。  不对，也不是特殊字符~~
        # return self.prettify(formatter=lambda s: s.replace(u'\xa0', ''))
        # return self.prettify()
        return str(self)

    def browser(self):
        browser.html(self)

    @run_once('id,str')
    def get_nonempty_childrens(self, *args):
        """ 获得所有Tag类型的直接子结点 （偏定制，不是那么通用的接口）

        会同时检查NavigableString类型，且必须是空白字符串，比如空格、\n之类
        """

        def check(x):
            if isinstance(x, bs4.element.Tag):
                return True
            elif isinstance(x, bs4.element.Comment):
                return False
            elif isinstance(x, bs4.element.NavigableString):
                assert not x.strip(), f'非空字符串值：{x}'
                return False
            else:
                raise ValueError(f'未见类型 {x}')

        ls = list(filter(check, self.children))

        if len(args):
            return ls[args[0]].get_nonempty_childrens(*args[1:])
        else:
            return ls

    def get_nonempty_children(self, *args):
        """ 输入args下标，指定获得某一个非空子结点 """
        if len(args):
            ls = self.get_nonempty_childrens(*args[:-1])
            return ls[args[-1]]
        else:
            return self

    def next_preorder_node(self, iter_child=True):
        """ 自己写的先序遍历

        主要应用在xml、bs4相关遍历检索时，有时候遇到特殊结点
            可能子结点不需要解析
            或者整个cur_node和子结点已经被解析完了，不需要再按照通常的先序遍历继续进入子结点
        此时可以 iter_child=False，进入下一个兄弟结点
        """
        # 传入的不一定是一个Tag结点~~
        if not isinstance(self, bs4.element.Tag):
            return None

        if iter_child and self.contents:
            return self.contents[0]
        else:
            cur_node = self
            while True:
                parent = cur_node.parent
                if parent is None:
                    return None
                sibing = cur_node.find_next_sibling()
                if sibing:
                    return sibing
                cur_node = parent

    def find_by_xpath(self, xpath):
        """ 使用xpath定位元素

        bs4官方没有自带，网上找到的很多也不中意。就自己根据需求简单定制一下。非完整版实现，但希望能支持常用的几个操作。
        好在还是有现成的xpath解析库的，自己扩展实现也不会太难。
        """
        from xpath_parser import XpathExpression

        xp = XpathExpression(xpath)

        cur_tag = self
        for node in xp.nodes:
            if node.name == '*':
                name = None
            else:
                name = node.name

            # TODO 其他前缀功能： .. 父结点， / 根节点
            recursive = node.ignore_position

            attrs = {}
            limit = 1
            for a in node.attrs:
                if a[0] == '@':
                    k, v = a.split('=')
                    attrs[k[1:]] = v[1:-1]
                elif re.match(r'\d+$', a):  # 索引下标
                    limit = int(a)
                else:
                    raise NotImplementedError

            # node.type没用上，应该有些需要用途的

            sub_tags = cur_tag.find_all(name, attrs, recursive, limit=limit)
            if sub_tags:
                cur_tag = sub_tags[-1]
            else:  # 没找到
                return None

        return cur_tag

    def __修改功能(self):
        pass

    @classmethod
    def _to_node(cls, html):
        """ 输入可以是字符串、文档、结点 """
        if isinstance(html, str):
            new_node = next(BeautifulSoup(html, 'lxml').body.children)
        elif html.find('body'):
            new_node = next(html.body.children)
        else:
            new_node = html
        return new_node

    @classmethod
    def _to_nodes(cls, html):
        """ 输入可以是字符串、文档、结点 """
        if isinstance(html, str):
            new_nodes = list(BeautifulSoup(html, 'lxml').body.children)
        elif html.find('body'):
            new_nodes = list(html.body.children)
        else:
            new_nodes = [html]
        return new_nodes

    def replace_html_with(self, html):
        nodes = self._to_nodes(html)  # 支持替换成多个节点
        if not nodes:
            return
        self.replace_with(nodes[0])

        cur = nodes[0]
        for node in nodes[1:]:
            cur.insert_after(node)
            cur = node

    def insert_html_before(self, html):
        nodes = self._to_nodes(html)
        if not nodes:
            return
        self.insert_before(nodes[0])

        cur = nodes[0]
        for node in nodes[1:]:
            cur.insert_after(node)
            cur = node

    def insert_html_after(self, html):
        nodes = self._to_nodes(html)
        if not nodes:
            return

        cur = self
        for node in nodes:
            cur.insert_after(node)
            cur = node

    def append_html(self, html):
        """ 原append的扩展 """
        nodes = self._to_nodes(html)
        for node in nodes:
            self.append(node)


inject_members(XlBs4Tag, bs4.element.Tag)
# 这样虽然不优雅，但主要是让特殊的String类型也支持兼容tag_name属性
inject_members(XlBs4Tag, bs4.element.NavigableString)


def mathjax_html_head(s):
    """增加mathjax解析脚本"""
    head = r"""<!DOCTYPE html>
<html>
<head>
<head><meta http-equiv=Content-Type content="text/html;charset=utf-8"></head>
<script src="https://a.cdn.histudy.com/lib/config/mathjax_config-klxx.js?v=1.1"></script>
<script type="text/javascript" async src="https://a.cdn.histudy.com/lib/mathjax/2.7.1/MathJax/MathJax.js?config=TeX-AMS-MML_SVG">
MathJax.Hub.Config(MATHJAX_KLXX_CONFIG);
</script>
</head>
<body>"""
    tail = '</body></html>'
    return head + s + tail


def html_bitran_template(htmlcontent):
    """ 双语翻译的html模板，html bilingual translation template

    一般是将word导出的html文件，转成方便谷歌翻译操作，进行双语对照的格式

    基本原理，是利用chrome识别class="notranslate"标记会跳过不翻译的特性
    对正文标签p拷贝两份，一份原文，一份带notranslate标记的内容
    这样在执行谷歌翻译后，就能出现双语对照的效果

    其实最好的办法，是能调用翻译API，直接给出双语成果的html
    但谷歌的googletrans连不上外网无法使用
    其他公司的翻译接口应该没问题，但我嫌其可能没有google好，以及不是重点，就先暂缓开发
    ---
    习惯来说，一般上面是英文，下面是中文，但是我又想使用中文标题~~
    """
    from pyxllib.text.nestenv import NestEnv

    # 0 将所有负margin-left变为0
    htmlcontent = re.sub(r'margin-left:-\d+(\.\d+)', 'margin-left:0', htmlcontent)

    # 1 区间定位分组
    ne = NestEnv(htmlcontent)
    ne2 = ne.xmltag('p')
    for name in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'pre', 'ol', 'li'):
        ne2 += ne.xmltag(name, symmetry=True)

    # 以下是针对python document复制到word的情况，不一定具有广泛泛用性
    # 目的是让代码块按块复制，而不是按行复制
    ne2 += ne.find2(re.compile("<div style=['\"]mso-element:para-border-div;.+?#AACC99"), '</div>')

    # 2 每个区间的处理规则
    def func(s):
        """ 找出p、h后，具体对每个tag要执行的操作

        分前后两波文本s1（原文），s2（翻译文）
        """

        # 1 s1 只要加 notranslate
        s1 = s
        bs = BeautifulSoup(s1, 'lxml')
        x = next(bs.body.children)
        cls_ = x.get('class', None)
        x['class'] = (cls_ + ['notranslate']) if cls_ else 'notranslate'
        s1 = x.prettify()

        # 2 s2 可能要做些骚操作
        s2 = s
        bs = BeautifulSoup(s2, 'lxml')
        x = next(bs.body.children)

        # 比如自定义翻译，这个无伤大雅的，如果搞不定，可以先注释掉，后面再说
        # if re.match(r'h\d+$', x.name):
        #     for y in x.descendants:
        #         if isinstance(y, NavigableString):
        #             y.replace_with(re.sub(r'Conclusion', '总结', str(y)))
        #         else:
        #             for z in y.strings:
        #                 z.replace_with(re.sub(r'Conclusion', '总结', str(z)))
        #     y.replace_with(re.sub(r'^Abstract$', '摘要', str(y)))
        # s2 = str(x)

        if re.match(r'h\d+$', x.name):
            x.name = 'p'  # 去掉标题格式，统一为段落格式
            s2 = x.prettify()
        elif x.name in ('div', 'pre'):
            # 实际使用体验，想了下，代码块还是不如保留原样最方便，不用拷贝翻译
            # s2 = x.prettify()
            s2 = ''  # 复制方式很有技巧
        # 如果p没有文本字符串，也不拷贝
        if not x.get_text().strip():
            s2 = ''
        # if x.name == 'p' and x.get('style', None) and 'margin-left' in x['style']:
        #     x['style'] = re.sub(r'(margin-left:)\d+(\.\d+)?', r'\g<1>0', x['style'])

        return s1 + '\n' + s2

    res = ne2.replace(func)

    return res


class MakeHtmlNavigation:
    """ 给网页添加一个带有超链接跳转的导航栏 """

    @classmethod
    def from_url(cls, url, **kwargs):
        """ 自动下载url的内容，缓存到本地后，加上导航栏打开 """
        content = requests.get(url).content.decode('utf8')
        etag = get_etag(url)  # 直接算url的etag，不用很严谨
        return cls.from_content(content, etag, **kwargs)

    @classmethod
    def from_file(cls, file, **kwargs):
        """ 输入本地一个html文件的路径，加上导航栏打开 """
        file = File(file)
        content = file.read()
        # 输入文件的情况，生成的_content等html要在同目录
        return cls.from_content(content, os.path.splitext(str(file))[0], **kwargs)

    @classmethod
    def from_content(cls, html_content, title='temphtml', *,
                     encoding=None, number=True, text_catalogue=True):
        """
        :param html_content: 原始网页的完整内容
        :param title: 页面标题，默认会先找head/title，如果没有，则取一个随机名称（TODO 未实装，目前固定名称）
        :param encoding: 保存的几个文件编码，默认是utf8，但windows平台有些特殊场合也可能要存储gbk
        :param number: 是否对每节启用自动编号的css

        算法基本原理：读取原网页，找出所有h标签，并增设a锚点
            另外生成一个导航html文件
            然后再生成一个主文件，让用户通过主文件来浏览页面

        # 读取csdn博客并展示目录 （不过因为这个存在跳级，效果不是那么好）
        >> file = 自动制作网页标题的导航栏(requests.get(r'https://blog.csdn.net/code4101/article/details/83009000').content.decode('utf8'))
        >> browser(str(file))
        http://i2.tiimg.com/582188/64f40d235705de69.png
        """
        from humanfriendly import format_size

        # 1 对原html，设置锚点，生成一个新的文件f2
        cnt = 0

        # 这个refs是可以用py算法生成的，目前是存储在github上引用
        refs = ['<html><head>',
                '<link rel=Stylesheet type="text/css" media=all '
                f'href="https://code4101.github.io/css/navigation{int(number)}.css">',
                '</head><body>']

        f2 = File(title + '_content', Dir.TEMP, suffix='.html')

        def func(m):
            nonlocal cnt
            cnt += 1
            name, content = m.group('name'), m.group('inner')
            content = BeautifulSoup(content, 'lxml').get_text()
            # 要写<h><a></a></h>，不能写<a><h></h></a>，否则css中设置的计数器重置不会起作用
            refs.append(f'<{name}><a href="{f2}#navigation{cnt}" target="showframe">{content}</a></{name}>')
            return f'<a name="navigation{cnt}"/>' + m.group()

        html_content = re.sub(r'<(?P<name>h\d+)(?:>|\s.*?>)(?P<body>\s*(?P<inner>.*?)\s*)</\1>',
                              func, html_content, flags=re.DOTALL)
        f2 = f2.write(html_content, encoding=encoding, if_exists='replace')

        # 2 f1除了导航栏，可以多附带一些有用的参考信息
        # 2.1 前文的refs已经存储了超链接的导航

        # 2.2 文本版的目录
        bs = BeautifulSoup(html_content, 'lxml')
        text = bs.get_text()
        if text_catalogue:
            # 目录
            refs.append(f'<br/>【文本版的目录】')
            catalogue = bs.get_catalogue(indent='\t', start_level=-1, jump=True, size=True)
            refs.append(f'<pre>{catalogue}</pre>')
            # 全文长度
            n = strwidth(text)
            refs.append('<br/>【Total Bytes】' + format_size(n))

        # 2.3 文中使用的高频词
        # 英文可以直接按空格切开统计，区分大小写
        text2 = re.sub(grp_chinese_char(), '', text)  # 删除中文，先不做中文的功能~~
        text2 = re.sub(r'[,\.，。\(\)（）;；?？"]', ' ', text2)  # 标点符号按空格处理
        words = Counter(text2.split())
        msg = '\n'.join([(x[0] if x[1] == 1 else f'{x[0]}，{x[1]}') for x in words.most_common()])
        msg += f'<br/>共{len(words)}个词汇，用词数{sum(words.values())}。'
        refs.append(f'<br/>【词汇表】<pre>{msg}</pre>')

        # 2.5 收尾，写入f1
        refs.append('</body>\n</html>')
        f1 = File(title + '_catalogue', Dir.TEMP, suffix='.html').write('\n'.join(refs), encoding=encoding,
                                                                        if_exists='replace')

        # 3 生成主页 f0
        main_content = f"""<html>
        <frameset cols="20%,80%">
        	<frame src="{f1}">
        	<frame src="{f2}" name="showframe">
        </frameset></html>"""

        f0 = File(title + '_index', Dir.TEMP, suffix='.html').write(main_content, encoding=encoding,
                                                                    if_exists='replace')
        return f0


class HtmlParser:
    """ 对树形结构、位置比较固定的html文档的一个解析框架 """

    def __init__(self, root):
        """ 输入根节点root """
        self.root = root

    @classmethod
    @run_once
    def get_parse_funcs(cls):
        res = []

        # 获取所有的方法名
        members = dir(cls)
        methods = filter(lambda m: callable(getattr(cls, m)), members)

        # 以parse、parse_0、parse_0_2等格式命名的函数，是解析树结构特定位置，这里自动执行解析
        for method in methods:
            if re.match(r'parse(_\d+)*$', method):
                # 智能获取对应下标的结构变量
                res.append(method)

        return res

    def run(self):
        for method in self.get_parse_funcs():
            # 智能获取对应下标的结构变量
            idxs = [int(v) for v in method[5:].split('_') if v]
            x = self.root.get_nonempty_children(*idxs)
            # 自动执行函数
            getattr(self, method)(x)


def concat_htmlbody(ls):
    """ 对多份网页内容中的body进行拼接
    """
    texts = [re.search(r'<body>(.*?)</body>', x, flags=re.DOTALL).group(1) for x in ls]
    # 用第一份作为主模板
    text = re.sub(r'<body>(.*?)</body>', lambda m: '<body>' + '\n'.join(texts) + '</body>', ls[0], flags=re.DOTALL)
    return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/03/28

import time

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import ElementClickInterceptedException, NoSuchWindowException


class element_has_text(object):
    def __init__(self, locator, text):
        self.locator = locator
        self.text = text

    def __call__(self, driver):
        element = driver.find_element(*self.locator)
        if self.text in element.text:
            return element
        else:
            return False


class XlChrome(webdriver.Chrome):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.maximize_window()

    def locate(self, locator, seconds=10):
        """ 定位一个元素 """
        if isinstance(locator, str):  # 默认是XPATH格式
            locator = (By.XPATH, locator)
        return WebDriverWait(self, seconds).until(EC.presence_of_element_located(locator))

    def click(self, locator, seconds=10, check=True):
        """ 点击一个元素 """
        if isinstance(locator, str):
            locator = (By.XPATH, locator)
        if check:
            element = WebDriverWait(self, seconds).until(EC.element_to_be_clickable(locator))
        else:
            element = self.locate(locator, seconds)
        time.sleep(0.5)  # 最好稍微等一下再点击
        try:
            element.click()
        except ElementClickInterceptedException:
            # 特殊情况，例如小鹅通下载页面的"下载"按钮没法正常click，要用js脚本去click
            self.execute_script("arguments[0].click();", element)

    def locate_text(self, locator, text, seconds=10):
        """ 判断指定元素位置是否含有指定文本 """
        if isinstance(locator, str):
            locator = (By.XPATH, locator)
        return WebDriverWait(self, seconds).until(element_has_text(locator, text))

    def __bool__(self):
        """ 判断driver是否还存在，如果已被手动关闭，这个值会返回False """
        try:
            self.title
            return True
        except NoSuchWindowException:
            return False


def get_global_driver(_driver_store=[None]):  # trick
    """ 通过这个接口可以固定一个driver来使用 """
    if _driver_store[0] is None:
        _driver_store[0] = XlChrome()
    if not _driver_store[0]:  # 如果驱动没了，重新启动
        _driver_store[0] = XlChrome()
    return _driver_store[0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/05/26

from pyxllib.prog.pupil import check_install_package

# 一个xpath解析库
check_install_package('jinja2')

import jinja2
from jinja2 import Template, Environment

from pyxllib.file.specialist import XlPath


def set_template(s, *args, **kwargs):
    """ todo 这个名字会不会太容易冲突了？ """
    return Template(s.strip(), *args, **kwargs)


def set_meta_template(s, meta_start='[[', meta_end=']]', **kwargs):
    """ 支持预先用某些格式渲染后，再返回标准渲染模板 """
    t = Template(s.strip(), variable_start_string=meta_start,
                 variable_end_string=meta_end).render(**kwargs)
    return Template(t)


def get_jinja_template(name, **kwargs):
    template = Environment(**kwargs).from_string((XlPath(__file__).parent / f'templates/{name}').read_text())
    return template



import os
import cv2
import numpy as np
from xlproject.code4101 import XlPath

from pyxllib.file.specialist import download_file
from pyxllib.prog.pupil import is_url


class SliderCaptchaLocator:
    """ 滑动验证码的位置定位

    由于缺口一般是一个黑底图，整体颜色是偏暗的，可以利用这个特性来定位列所在位置
    """

    def __init__(self, background_image_path, slider_image_path=None):
        """
        :param background_image_path: 背景主图片
        :param slider_image_path: 滑动条图片
        """
        # 支持输入是url的情况
        if is_url(background_image_path):
            background_image_path = download_file(background_image_path, temp=True, ext='.png')
        if is_url(slider_image_path):
            slider_image_path = download_file(slider_image_path, temp=True, ext='.png')
        # 初始化图片等数据
        self.image_path = background_image_path
        self.image = cv2.imread(self.image_path)
        self.darkest_rows = None
        self.column_averages = None
        self.enhanced_averages = None
        self.radius = 24
        if slider_image_path is not None:
            # 先读取图片
            template_image = cv2.imread(slider_image_path)
            # 取图片宽度的40%作为radius
            self.radius = int(template_image.shape[1] * 0.4)

    def select_darkest_part_rows(self, fraction=0.1):
        """
        选择最暗的部分行，供下游计算列的平均暗度

        :param float fraction: 最暗行的比例
        :return: 最暗行的图像部分
        """
        if self.darkest_rows is None:
            average_row_values = np.mean(self.image, axis=(1, 2))
            threshold_index = int(len(average_row_values) * fraction)
            darkest_indices = np.argsort(average_row_values)[:threshold_index]
            self.darkest_rows = self.image[darkest_indices]
        return self.darkest_rows

    def calculate_column_averages(self):
        """
        计算列的平均强度。

        :return: 列的平均强度
        """
        if self.column_averages is None:
            rows = self.select_darkest_part_rows()
            column_values = np.mean(rows, axis=0)
            column_intensity = np.mean(column_values, axis=1)
            min_val, max_val = np.min(column_intensity), np.max(column_intensity)
            self.column_averages = ((column_intensity - min_val) / (max_val - min_val) * 255).astype(
                np.uint8) if max_val != min_val else np.zeros_like(column_intensity)
        return self.column_averages

    def enhance_column_averages(self):
        """
        增强列平均值，利用验证码的黑度图的对称性，进行一个增强，使得其他非验证码的零散黑条变白

        :param int radius: 对比半径，这个建议设成模版图宽度的40%~50%
        :return: 增强后的列平均值
        """
        if self.enhanced_averages is None:
            averages = self.calculate_column_averages()
            length = len(averages)
            enhanced = np.zeros_like(averages)
            for i in range(length):
                max_value = averages[i]
                for j in range(1, self.radius // 2 + 1):
                    if i - j >= 0 and i + j < length:
                        max_value = max(max_value, abs(int(averages[i - j]) - int(averages[i + j])))
                enhanced[i] = max_value
            self.enhanced_averages = enhanced
        return self.enhanced_averages

    def find_captcha_position(self, threshold=50):
        """
        查找验证码的位置。

        :param int threshold: 达到多少像素后，开始计算中位数
        :return: 验证码的列中心位置
        """
        intensities = self.enhance_column_averages()
        indices = []
        for intensity in range(256):
            new_indices = np.where(intensities == intensity)[0]
            if len(indices) + len(new_indices) > threshold:
                indices.extend(new_indices[:threshold - len(indices)])
                break
            indices.extend(new_indices)
        return int(np.median(indices))

    def draw_red_line(self, position, line_thickness=3):
        """
        在图像上绘制红线。

        :param int position: 红线位置
        :param int line_thickness: 线宽
        :return: 绘制红线后的图像
        """
        gray_image = np.tile(self.calculate_column_averages(), (100, 1)).astype(np.uint8)
        color_image = cv2.cvtColor(gray_image, cv2.COLOR_GRAY2BGR)
        cv2.line(color_image, (position, 0), (position, gray_image.shape[0]), (0, 0, 255), line_thickness)
        return color_image

    def debug(self):
        position = self.find_captcha_position()
        final_image = self.draw_red_line(position)
        cv2.imshow('Image with Red Line', final_image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()


def main():
    os.chdir(XlPath.desktop())
    locator = SliderCaptchaLocator('cap_union_new_getcapbysig.png')
    position = locator.find_captcha_position()
    final_image = locator.draw_red_line(position)
    cv2.imshow('Image with Red Line', final_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/11/19

""" 问卷星 相关工具 """

import os
import io
import time

from loguru import logger
from DrissionPage import Chromium
import pandas as pd

from pyxllib.ext.drissionlib import DpWebBase


class WjxWeb(DpWebBase):
    """ 问卷星网页的爬虫 """

    def __init__(self, url=None):
        super().__init__(url or 'https://www.wjx.cn')
        self.login()

    def login(self):
        tab = self.tab

        if tab.url.startswith('https://www.wjx.cn/wjx/activitystat/resultlimit.aspx'):
            tab('tag:a@@text():登录').click()

        if tab.url.lower().startswith('https://www.wjx.cn/login.aspx'):
            tab('tag:input@@name=UserName').input(os.getenv('WJX_USERNAME'), clear=True)
            tab('tag:input@@name=Password').input(os.getenv('WJX_PASSWORD'), clear=True)
            tab('tag:input@@type=submit').click()

    def get_page_num(self):
        """
        返回当前页编号和总页数 (idx, num)。
        """
        idx, num = map(int, self.tab('tag:span@@class=paging-num').text.split('/'))
        return idx, num

    def prev_page(self):
        self.tab('tag:a@@class=go-pre').click()

    def next_page(self):
        self.tab('tag:a@@class=go-next').click()

    def _parse_table(self):
        """处理并解析网页中的表格数据"""
        table_html = self.tab('tag:table').html
        df = pd.read_html(io.StringIO(table_html))[0]  # 读取表格
        df.columns = [col.replace('\ue645', '') for col in df.columns]
        # "星标"的内容特殊字符
        df.replace('\ue66b', '', regex=True, inplace=True)
        # "操作"的内容特殊字符
        df.replace('\ue6a3\ue6d4', '', regex=True, inplace=True)
        return df

    def set_num_of_page(self, num_of_page):
        """ 查看数据页面，设置每页显示多少条记录 """
        select = self.tab('tag:span@@text():每页显示').next('tag:select')
        select.click()
        opt = select(f'tag:option@@text()={num_of_page}')
        if opt.attr('selected') != 'selected':
            opt.click()
        else:
            select.click()

    def get_df(self, all_pages=False):
        """获得当前页面的表格数据，如果 all_pages 为 True，则下载所有页面的数据"""
        # 初始化DataFrame列表，用于存储每页的数据
        dfs = [self._parse_table()]  # 获取当前页面的数据

        # 如果需要下载所有页面数据
        if all_pages:
            current_idx, total_pages = self.get_page_num()
            while current_idx < total_pages:
                self.next_page()  # 翻到下一页
                time.sleep(2)
                dfs.append(self._parse_table())  # 获取并处理新一页的数据
                current_idx, total_pages = self.get_page_num()  # 更新页码信息

        # 将所有数据合并为一个DataFrame
        final_df = pd.concat(dfs, ignore_index=True) if all_pages else dfs[0]
        return final_df



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import paddle
from paddle import nn
import paddle.nn.functional as F
from paddle import ParamAttr


class ConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 stride=1,
                 groups=1,
                 is_vd_mode=False,
                 act=None,
                 name=None):
        super(ConvBNLayer, self).__init__()

        self.is_vd_mode = is_vd_mode
        self._pool2d_avg = nn.AvgPool2D(
            kernel_size=2, stride=2, padding=0, ceil_mode=True)
        self._conv = nn.Conv2D(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2,
            groups=groups,
            weight_attr=ParamAttr(name=name + "_weights"),
            bias_attr=False)
        if name == "conv1":
            bn_name = "bn_" + name
        else:
            bn_name = "bn" + name[3:]
        self._batch_norm = nn.BatchNorm(
            out_channels,
            act=act,
            param_attr=ParamAttr(name=bn_name + '_scale'),
            bias_attr=ParamAttr(bn_name + '_offset'),
            moving_mean_name=bn_name + '_mean',
            moving_variance_name=bn_name + '_variance',
            use_global_stats=False)

    def forward(self, inputs):
        y = self._conv(inputs)
        y = self._batch_norm(y)
        return y


class DeConvBNLayer(nn.Layer):
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size=4,
                 stride=2,
                 padding=1,
                 groups=1,
                 if_act=True,
                 act=None,
                 name=None):
        super(DeConvBNLayer, self).__init__()

        self.if_act = if_act
        self.act = act
        self.deconv = nn.Conv2DTranspose(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=groups,
            weight_attr=ParamAttr(name=name + '_weights'),
            bias_attr=False)
        self.bn = nn.BatchNorm(
            num_channels=out_channels,
            act=act,
            param_attr=ParamAttr(name="bn_" + name + "_scale"),
            bias_attr=ParamAttr(name="bn_" + name + "_offset"),
            moving_mean_name="bn_" + name + "_mean",
            moving_variance_name="bn_" + name + "_variance",
            use_global_stats=False)

    def forward(self, x):
        x = self.deconv(x)
        x = self.bn(x)
        return x


class PGFPN(nn.Layer):
    def __init__(self, in_channels, **kwargs):
        super(PGFPN, self).__init__()
        num_inputs = [2048, 2048, 1024, 512, 256]
        num_outputs = [256, 256, 192, 192, 128]
        self.out_channels = 128
        self.conv_bn_layer_1 = ConvBNLayer(
            in_channels=3,
            out_channels=32,
            kernel_size=3,
            stride=1,
            act=None,
            name='FPN_d1')
        self.conv_bn_layer_2 = ConvBNLayer(
            in_channels=64,
            out_channels=64,
            kernel_size=3,
            stride=1,
            act=None,
            name='FPN_d2')
        self.conv_bn_layer_3 = ConvBNLayer(
            in_channels=256,
            out_channels=128,
            kernel_size=3,
            stride=1,
            act=None,
            name='FPN_d3')
        self.conv_bn_layer_4 = ConvBNLayer(
            in_channels=32,
            out_channels=64,
            kernel_size=3,
            stride=2,
            act=None,
            name='FPN_d4')
        self.conv_bn_layer_5 = ConvBNLayer(
            in_channels=64,
            out_channels=64,
            kernel_size=3,
            stride=1,
            act='relu',
            name='FPN_d5')
        self.conv_bn_layer_6 = ConvBNLayer(
            in_channels=64,
            out_channels=128,
            kernel_size=3,
            stride=2,
            act=None,
            name='FPN_d6')
        self.conv_bn_layer_7 = ConvBNLayer(
            in_channels=128,
            out_channels=128,
            kernel_size=3,
            stride=1,
            act='relu',
            name='FPN_d7')
        self.conv_bn_layer_8 = ConvBNLayer(
            in_channels=128,
            out_channels=128,
            kernel_size=1,
            stride=1,
            act=None,
            name='FPN_d8')

        self.conv_h0 = ConvBNLayer(
            in_channels=num_inputs[0],
            out_channels=num_outputs[0],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_h{}".format(0))
        self.conv_h1 = ConvBNLayer(
            in_channels=num_inputs[1],
            out_channels=num_outputs[1],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_h{}".format(1))
        self.conv_h2 = ConvBNLayer(
            in_channels=num_inputs[2],
            out_channels=num_outputs[2],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_h{}".format(2))
        self.conv_h3 = ConvBNLayer(
            in_channels=num_inputs[3],
            out_channels=num_outputs[3],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_h{}".format(3))
        self.conv_h4 = ConvBNLayer(
            in_channels=num_inputs[4],
            out_channels=num_outputs[4],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_h{}".format(4))

        self.dconv0 = DeConvBNLayer(
            in_channels=num_outputs[0],
            out_channels=num_outputs[0 + 1],
            name="dconv_{}".format(0))
        self.dconv1 = DeConvBNLayer(
            in_channels=num_outputs[1],
            out_channels=num_outputs[1 + 1],
            act=None,
            name="dconv_{}".format(1))
        self.dconv2 = DeConvBNLayer(
            in_channels=num_outputs[2],
            out_channels=num_outputs[2 + 1],
            act=None,
            name="dconv_{}".format(2))
        self.dconv3 = DeConvBNLayer(
            in_channels=num_outputs[3],
            out_channels=num_outputs[3 + 1],
            act=None,
            name="dconv_{}".format(3))
        self.conv_g1 = ConvBNLayer(
            in_channels=num_outputs[1],
            out_channels=num_outputs[1],
            kernel_size=3,
            stride=1,
            act='relu',
            name="conv_g{}".format(1))
        self.conv_g2 = ConvBNLayer(
            in_channels=num_outputs[2],
            out_channels=num_outputs[2],
            kernel_size=3,
            stride=1,
            act='relu',
            name="conv_g{}".format(2))
        self.conv_g3 = ConvBNLayer(
            in_channels=num_outputs[3],
            out_channels=num_outputs[3],
            kernel_size=3,
            stride=1,
            act='relu',
            name="conv_g{}".format(3))
        self.conv_g4 = ConvBNLayer(
            in_channels=num_outputs[4],
            out_channels=num_outputs[4],
            kernel_size=3,
            stride=1,
            act='relu',
            name="conv_g{}".format(4))
        self.convf = ConvBNLayer(
            in_channels=num_outputs[4],
            out_channels=num_outputs[4],
            kernel_size=1,
            stride=1,
            act=None,
            name="conv_f{}".format(4))

    def forward(self, x):
        c0, c1, c2, c3, c4, c5, c6 = x
        # FPN_Down_Fusion
        f = [c0, c1, c2]
        g = [None, None, None]
        h = [None, None, None]
        h[0] = self.conv_bn_layer_1(f[0])
        h[1] = self.conv_bn_layer_2(f[1])
        h[2] = self.conv_bn_layer_3(f[2])

        g[0] = self.conv_bn_layer_4(h[0])
        g[1] = paddle.add(g[0], h[1])
        g[1] = F.relu(g[1])
        g[1] = self.conv_bn_layer_5(g[1])
        g[1] = self.conv_bn_layer_6(g[1])

        g[2] = paddle.add(g[1], h[2])
        g[2] = F.relu(g[2])
        g[2] = self.conv_bn_layer_7(g[2])
        f_down = self.conv_bn_layer_8(g[2])

        # FPN UP Fusion
        f1 = [c6, c5, c4, c3, c2]
        g = [None, None, None, None, None]
        h = [None, None, None, None, None]
        h[0] = self.conv_h0(f1[0])
        h[1] = self.conv_h1(f1[1])
        h[2] = self.conv_h2(f1[2])
        h[3] = self.conv_h3(f1[3])
        h[4] = self.conv_h4(f1[4])

        g[0] = self.dconv0(h[0])
        g[1] = paddle.add(g[0], h[1])
        g[1] = F.relu(g[1])
        g[1] = self.conv_g1(g[1])
        g[1] = self.dconv1(g[1])

        g[2] = paddle.add(g[1], h[2])
        g[2] = F.relu(g[2])
        g[2] = self.conv_g2(g[2])
        g[2] = self.dconv2(g[2])

        g[3] = paddle.add(g[2], h[3])
        g[3] = F.relu(g[3])
        g[3] = self.conv_g3(g[3])
        g[3] = self.dconv3(g[3])

        g[4] = paddle.add(x=g[3], y=h[4])
        g[4] = F.relu(g[4])
        g[4] = self.conv_g4(g[4])
        f_up = self.convf(g[4])
        f_common = paddle.add(f_down, f_up)
        f_common = F.relu(f_common)
        return f_common



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/10/29

import platform
import sys
import time

import cv2
import psutil
import numpy as np

if sys.platform == 'win32':
    import win32gui
    import win32process
    # from pywinauto import Desktop
    import uiautomation

from pyxllib.prog.cachetools import xlcache
from pyxllib.cv.xlcvlib import CvImg
from pyxlpr.ai.clientlib import XlAiClient

# 根据平台扩展相应的类
if platform.system().lower() == "darwin":
    from mss.darwin import MSS as OriginalMSS
elif platform.system().lower() == "linux":
    from mss.linux import MSS as OriginalMSS
elif platform.system().lower() == "windows":
    from mss.windows import MSS as OriginalMSS
else:
    raise RuntimeError("Unsupported platform")


def adjust_monitors(monitors, scale_factor):
    # 调整所有物理显示器的尺寸和位置
    for i in range(1, len(monitors)):
        monitors[i]['width'] = int(monitors[i]['width'] / scale_factor)
        monitors[i]['height'] = int(monitors[i]['height'] / scale_factor)
        monitors[i]['top'] = int(monitors[i]['top'] / scale_factor)
        monitors[i]['left'] = int(monitors[i]['left'] / scale_factor)

    # 计算虚拟桌面的边界
    left_bound = min(monitor['left'] for monitor in monitors[1:])
    top_bound = min(monitor['top'] for monitor in monitors[1:])
    right_bound = max(monitor['left'] + monitor['width'] for monitor in monitors[1:])
    bottom_bound = max(monitor['top'] + monitor['height'] for monitor in monitors[1:])

    # 设置虚拟桌面的尺寸和位置
    monitors[0]['width'] = right_bound - left_bound
    monitors[0]['height'] = bottom_bound - top_bound
    monitors[0]['top'] = top_bound
    monitors[0]['left'] = left_bound

    return monitors


class ActiveWindowCapture(OriginalMSS):
    """
    基于`MSS`库的一个扩展，用于实现跨平台的多屏幕截图，并添加了以下独特功能：
    1. 实时窗口捕获：能够获取当前激活的窗口信息，并关联到对应的进程对象。
        这一特性使得该类在需要实时监控、分析特定应用窗口时尤为实用。
    2. 缓存机制：该类采用了缓存策略，通过`xlcache`装饰器缓存截取的屏幕和窗口数据，避免频繁重复捕获，提升性能。
        缓存机制同时确保在多窗口环境中，当前获取的窗口信息是实际的活动窗口，防止数据滞后或错位。
    3. 多屏幕拼接与坐标调整：支持多屏幕拼接功能，能够调整不同显示器的坐标和尺寸，以生成全屏拼接的截图画布。
        对于分辨率缩放和多屏偏移，也可以进行自适应调整。
    4. 标记窗口功能：在截屏时，可以选择性地标记当前激活窗口的位置，支持自定义标记的颜色、厚度和偏移量，便于在截图中直观地显示当前活跃窗口。

    适合需要频繁截图、实时监控活动窗口和对应进程的场景，尤其在多屏和分辨率适配方面提供了便捷的解决方案。
    """

    def __1_动态成员变量(self):
        pass

    def __init__(self, *, scale=1):
        """
        :param scale: 可适当兼容主屏幕做了缩放处理的情景，但太复杂的情况下也不一定对
        """
        super().__init__()
        self.scale = scale

    @xlcache(property=True)
    def monitors2(self):
        """ 修正后的各个屏幕坐标 """
        monitors = self.monitors
        if self.scale != 1:
            monitors = adjust_monitors(monitors, self.scale)
        return monitors

    @xlcache(property=True)
    def window(self):
        """ 当前激活窗口 """
        # 可以小等一会等到有效窗口
        for i in range(4):
            win = uiautomation.GetForegroundControl()
            if win:
                return win
            else:
                time.sleep(0.5)
        return

    @xlcache(property=True)
    def process(self):
        """ 当前激活窗口 """
        hwnd = self.window.NativeWindowHandle
        thread_id, process_id = win32process.GetWindowThreadProcessId(hwnd)
        return psutil.Process(process_id)

    @xlcache(property=True)
    def xlapi(self):
        return XlAiClient()

    def __2_截图数据(self):
        pass

    @xlcache()
    def _capture_single_monitor(self, order):
        """ 缓存单个屏幕的截图 """
        if order == 0:
            # order=0时，拼接所有屏幕图像。虽然mss其实可以直接提供，但是这样就没有每个屏幕缓存的图片了，还是我自己处理一遍更好。

            # 计算坐标偏移量，将所有坐标调整至非负
            min_left = min(monitor['left'] for monitor in self.monitors2)
            min_top = min(monitor['top'] for monitor in self.monitors2)

            # 获取总画布大小（根据所有屏幕的范围计算）
            max_right = max(monitor['left'] + monitor['width'] for monitor in self.monitors2)
            max_bottom = max(monitor['top'] + monitor['height'] for monitor in self.monitors2)
            canvas_width = max_right - min_left
            canvas_height = max_bottom - min_top

            # 使用numpy创建空白画布 (BGR格式)
            img_canvas = np.zeros((canvas_height, canvas_width, 3), dtype=np.uint8)

            # 获取并缓存每个子屏幕截图，按调整后的坐标拼接到总画布
            for i in range(1, len(self.monitors2)):
                monitor = self.monitors2[i]
                single_img = self._capture_single_monitor(i)  # 获取单屏截图 (cv2格式)
                # 将子屏幕截图粘贴到总画布的指定位置，应用坐标偏移
                x_offset = monitor['left'] - min_left
                y_offset = monitor['top'] - min_top
                img_canvas[y_offset:y_offset + single_img.shape[0],
                x_offset:x_offset + single_img.shape[1]] = single_img

            img_cv2 = img_canvas
        else:
            if order == -1:
                # 截取当前活动窗口的截图
                active_window = self.window
                rect = active_window.BoundingRectangle
                monitor = {
                    'left': int(rect.left),
                    'top': int(rect.top),
                    'width': int(rect.right - rect.left),
                    'height': int(rect.bottom - rect.top),
                }
                sct_img = self.grab(monitor)
            else:
                sct_img = self.grab(self.monitors2[order])

            # 将截图的原始数据转换成numpy数组
            img_np = np.frombuffer(sct_img.bgra, dtype=np.uint8)
            img_np = img_np.reshape((sct_img.height, sct_img.width, 4))  # 高、宽、4通道
            # 转换成BGR通道顺序
            img_cv2 = cv2.cvtColor(img_np, cv2.COLOR_BGRA2BGR)

        return img_cv2

    def _mark_window_on_screenshot(self, screenshot, window_rect, order=0, mark_params=None):
        """
        在截图中标记窗口，并根据 order 参数自动调整坐标偏移。

        :param screenshot: 要标记的截图图像
        :param window_rect: 当前激活窗口的原始坐标 (left, top, right, bottom)
        :param order: 屏幕索引 (0 表示拼接所有屏幕，全图模式)
        :param mark_params: 标记参数字典，包含 color, base, thickness 等
        :return: 标记后的截图
        """
        # 1 设定默认标记参数，若传入 mark_params 则进行更新
        default_params = {
            "color": (0, 255, 0),  # 默认绿色
            "base": 10,  # 默认缩进 10 像素
            "thickness": 3  # 默认厚度 3 像素
        }
        mark_params = mark_params if isinstance(mark_params, dict) else None
        if mark_params:
            default_params.update(mark_params)

        color = default_params["color"]
        base = default_params["base"]
        thickness = default_params["thickness"]

        # 2 获取对应 order 的屏幕原点坐标进行偏差修正
        if order == 0:
            origin_x = self.monitors2[0]['left']
            origin_y = self.monitors2[0]['top']
        else:
            origin_x = self.monitors2[order]['left']
            origin_y = self.monitors2[order]['top']

        # 计算相对于指定屏幕或全图的窗口坐标
        adjusted_coords = (
            int(window_rect.left) - origin_x,
            int(window_rect.top) - origin_y,
            int(window_rect.right) - origin_x,
            int(window_rect.bottom) - origin_y,
        )

        # 3 绘制矩形标记
        cv2.rectangle(
            screenshot,
            (adjusted_coords[0] + base, adjusted_coords[1] + base),
            (adjusted_coords[2] - base, adjusted_coords[3] - base),
            color,  # 使用自定义颜色
            thickness  # 使用自定义厚度
        )

        return screenshot

    @xlcache()
    def screenshot(self, order=0, *, to_pil=False, mark_active_window=False):
        """ 截屏

        :param int order:
            -1, 截取当前激活窗口
            0，默认值，截取全部屏幕
            1，截取第1个屏幕
            2，截取第2个屏幕
            ...
        :param to_pil: 是否转换成 PIL 格式
        :param mark_active_window: 是否标记当前激活窗口，若为字典可设置颜色、厚度等参数
        :return: np矩阵或PIL图片，取决于to_cv2参数
        """
        assert order < len(self.monitors), '屏幕下标越界'
        img = self._capture_single_monitor(order)  # 调用子函数处理所有屏幕拼接

        # 若需要标记当前激活窗口
        if mark_active_window:
            img = self._mark_window_on_screenshot(img, self.window.BoundingRectangle, order,
                                                  mark_params=mark_active_window)

        if to_pil:
            img = CvImg.read(img).to_pil_image()  # 转成pil矩阵格式

        return img



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/WenmuZhou/DBNet.pytorch/blob/master/data_loader/modules/make_shrink_map.py
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import cv2
from shapely.geometry import Polygon
import pyclipper

__all__ = ['MakeShrinkMap']


class MakeShrinkMap(object):
    r'''
    Making binary mask from detection data with ICDAR format.
    Typically following the process of class `MakeICDARData`.
    '''

    def __init__(self, min_text_size=8, shrink_ratio=0.4, **kwargs):
        self.min_text_size = min_text_size
        self.shrink_ratio = shrink_ratio

    def __call__(self, data):
        image = data['image']
        text_polys = data['polys']
        ignore_tags = data['ignore_tags']

        h, w = image.shape[:2]
        # 1 校验文本检测标签
        text_polys, ignore_tags = self.validate_polygons(text_polys,
                                                         ignore_tags, h, w)
        gt = np.zeros((h, w), dtype=np.float32)
        mask = np.ones((h, w), dtype=np.float32)
        # 2 根据文本检测框计算文本区域概率图
        for i in range(len(text_polys)):
            polygon = text_polys[i]
            height = max(polygon[:, 1]) - min(polygon[:, 1])
            width = max(polygon[:, 0]) - min(polygon[:, 0])
            if ignore_tags[i] or min(height, width) < self.min_text_size:
                cv2.fillPoly(mask,
                             polygon.astype(np.int32)[np.newaxis, :, :], 0)
                ignore_tags[i] = True
            else:
                # 多边形内缩
                polygon_shape = Polygon(polygon)
                subject = [tuple(l) for l in polygon]
                padding = pyclipper.PyclipperOffset()
                padding.AddPath(subject, pyclipper.JT_ROUND,
                                pyclipper.ET_CLOSEDPOLYGON)
                shrinked = []

                # Increase the shrink ratio every time we get multiple polygon returned back
                possible_ratios = np.arange(self.shrink_ratio, 1,
                                            self.shrink_ratio)
                np.append(possible_ratios, 1)
                # print(possible_ratios)
                for ratio in possible_ratios:
                    # print(f"Change shrink ratio to {ratio}")
                    distance = polygon_shape.area * (
                        1 - np.power(ratio, 2)) / polygon_shape.length
                    shrinked = padding.Execute(-distance)
                    if len(shrinked) == 1:
                        break

                if shrinked == []:
                    cv2.fillPoly(mask,
                                 polygon.astype(np.int32)[np.newaxis, :, :], 0)
                    ignore_tags[i] = True
                    continue

                # 填充
                for each_shirnk in shrinked:
                    shirnk = np.array(each_shirnk).reshape(-1, 2)
                    cv2.fillPoly(gt, [shirnk.astype(np.int32)], 1)

        data['shrink_map'] = gt
        data['shrink_mask'] = mask
        return data

    def validate_polygons(self, polygons, ignore_tags, h, w):
        '''
        polygons (numpy.array, required): of shape (num_instances, num_points, 2)
        '''
        if len(polygons) == 0:
            return polygons, ignore_tags
        assert len(polygons) == len(ignore_tags)
        for polygon in polygons:
            polygon[:, 0] = np.clip(polygon[:, 0], 0, w - 1)
            polygon[:, 1] = np.clip(polygon[:, 1], 0, h - 1)

        for i in range(len(polygons)):
            area = self.polygon_area(polygons[i])
            if abs(area) < 1:
                ignore_tags[i] = True
            if area > 0:
                polygons[i] = polygons[i][::-1, :]
        return polygons, ignore_tags

    def polygon_area(self, polygon):
        """
        compute polygon area
        """
        area = 0
        q = polygon[-1]
        for p in polygon:
            area += p[0] * q[1] - p[1] * q[0]
            q = p
        return area / 2.0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/04/27

"""
libreoffice这个软件相关的功能

官网：https://www.libreoffice.org/download/download-libreoffice/

linux的安装：
sudo apt-get update
sudo apt-get install libreoffice -y
"""

import os
from pathlib import Path
import subprocess
import sys
import tempfile
from datetime import datetime


def check_libreoffice():
    """ 检查LibreOffice是否安装 """
    executor = get_libreoffice_executor()
    try:
        subprocess.run([executor, '--version'], check=True)
    except FileNotFoundError:
        return False
    except subprocess.CalledProcessError:
        return False
    return True


def get_libreoffice_executor():
    """ 获得可执行文件的名称 """
    # 根据所在系统区分
    if sys.platform == 'win32':
        return 'soffice.exe'
    else:
        return 'libreoffice'


def infer_file_format(file_path):
    """ 推断文件所属办公文档的'主类' """
    ext = Path(file_path).suffix.lower()
    if ext in ['.doc', '.docx']:
        fmt = 'docx'
    elif ext in ['.ppt', '.pptx']:
        fmt = 'pptx'
    elif ext in ['.xls', '.xlsx']:
        # todo 其实excel还有xlsm的可能，但是用libreoffice可能没有微软365那样可以判断出宏的情况保存为xlsm
        fmt = 'xlsx'
    else:
        raise ValueError("不支持的文件格式")
    return fmt


class UpgradeOfficeFile:
    @classmethod
    def to_dir(cls, file_path, out_dir=None, fmt=None, timeout=10):
        """ 将doc文件转换为docx文件

        :param file_path: 待升级的文件路径
        :param out_dir: 输出文件目录
            官方接口默认只能设置导出目录，不能设置导出文件名，文件名是跟原始文件一样的
        :param fmt: 输出文件格式
            docx, xlsx, pptx
        """
        if isinstance(file_path, Path):
            file_path = file_path.as_posix()

        # 获取LibreOffice可执行文件的名称
        executor = get_libreoffice_executor()

        # 如果未指定输出路径，则默认在输入文件的同目录下生成同名的DOCX文件
        if out_dir is None:
            out_dir = os.path.dirname(file_path)

        if fmt is None:
            fmt = infer_file_format(file_path)

        # 构建转换命令
        command = [
            executor,
            '--headless',  # 无界面模式
            '--convert-to', fmt,  # 转换为docx格式
            '--outdir', str(out_dir),  # 输出目录
            file_path  # 输入文件路径
        ]

        subprocess.run(command, timeout=timeout, check=True)

        # 返回转换后的文件路径
        base_name = os.path.basename(file_path)
        name, _ = os.path.splitext(base_name)
        new_file_path = os.path.join(out_dir, f"{name}.{fmt}")

        # todo 以目标文件是否存在判断转换是否成功也是有一定bug的，可能目标文件本来就存在
        #   但如果严谨判断，就要分析subprocess.run的输出结果了，那个太麻烦，先用简便方法处理
        if not Path(new_file_path).exists():
            raise ValueError(f"升级文档失败")

        return new_file_path

    @classmethod
    def to_file(cls, in_file, out_file=None, fmt=None, timeout=10):
        """ 可以指定转换出的文件名的版本

        :param in_file: 待转换的文件路径
        :param out_file: 输出文件路径
            若未指定,默认与原文件同名,在原文件所在目录生成
        :param fmt: 输出文件格式
            docx, xlsx, pptx
            若未指定,则根据in_file的后缀名自动判断
        """
        # 将in_file转换为Path对象
        in_file = Path(in_file)

        # 如果fmt为None,则根据in_file推断
        if fmt is None:
            fmt = infer_file_format(in_file)

        # 如果out_file为None,则默认在原文件目录生成同名的新格式文件
        if out_file is None:
            out_file = in_file.with_suffix(f'.{fmt}')
        else:
            out_file = Path(out_file)

        # 确保out_file的父目录存在,不存在则创建
        out_file.parent.mkdir(parents=True, exist_ok=True)

        # 调用upgrade_office_file函数进行转换
        temp_file = cls.to_dir(in_file, out_dir=out_file.parent, fmt=fmt, timeout=timeout)

        # 将生成的临时文件重命名为out_file
        os.rename(temp_file, out_file)

        return out_file

    @classmethod
    def to_tempfile(cls, in_file, fmt=None, *, timestamp_stem=False, create_subdir=False, timeout=10):
        """ 将文件转换为临时文件

        :param timestamp_stem: 时间戳文件名
        :param create_subdir: 是否在临时目录中创建新的子目录
        """
        if fmt is None:
            fmt = infer_file_format(in_file)

        root = Path(tempfile.gettempdir())
        if create_subdir:
            root2 = root / datetime.now().strftime('%Y%m%d.%H%M%S.%f')
            root2.mkdir(parents=True, exist_ok=True)
            root = root2

        if timestamp_stem:
            stem = datetime.now().strftime('%Y%m%d.%H%M%S.%f')
            out_file = cls.to_file(in_file, out_file=root / f"{stem}.{fmt}", fmt=fmt, timeout=timeout)
        else:
            out_file = cls.to_dir(in_file, out_dir=root, fmt=fmt, timeout=timeout)

        return out_file



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/10/30


import os
import json
import base64

from pyxllib.text.newbie import add_quote


class XlOsEnv:
    """ pyxllib库自带的一套环境变量数据解析类

    会将json的字符串值，或者普通str，存储到环境变量中

    环境变量也可以用来实现全局变量的信息传递，虽然不太建议这样做

    >> XlOsEnv.persist_set('TP10_ACCOUNT',
                           {'server': '172.16.250.250', 'port': 22, 'user': 'ckz', 'passwd': '123456'},
                           True)
    >> print(XlOsEnv.get('TP10_ACCOUNT'), True)  # 展示存储的账号信息
    eyJzZXJ2ZXIiOiAiMTcyLjE2LjE3MC4xMzQiLCAicG9ydCI6IDIyLCAidXNlciI6ICJjaGVua3VuemUiLCAicGFzc3dkIjogImNvZGV4bHByIn0=
    >> XlOsEnv.unset('TP10_ACCOUNT')
    """

    @classmethod
    def get(cls, name, *, decoding=False):
        """ 获取环境变量值

        :param name: 环境变量名
        :param decoding: 是否需要先进行base64解码
        :return:
            返回json解析后的数据
            或者普通的字符串值
        """
        value = os.getenv(name, None)
        if value is None:
            return value

        if decoding:
            value = base64.b64decode(value.encode())

        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            if isinstance(value, bytes):
                return value.decode()
            else:
                return value

    @classmethod
    def set(cls, name, value, encoding=False):
        """ 临时改变环境变量

        :param name: 环境变量名
        :param value: 要存储的值
        :param encoding: 是否将内容转成base64后，再存储环境变量
            防止一些密码信息，明文写出来太容易泄露
            不过这个策略也很容易被破解；只防君子，难防小人

            当然，谁看到这有闲情功夫的话，可以考虑做一套更复杂的加密系统
            并且encoding支持多种不同的解加密策略，这样单看环境变量值就很难破译了
        :return: str, 最终存储的字符串内容
        """
        # 1 打包
        if isinstance(value, str):
            value = add_quote(value)
        else:
            value = json.dumps(value)

        # 2 编码
        if encoding:
            value = base64.b64encode(value.encode()).decode()

        # 3 存储到环境变量
        os.environ[name] = value

        return value

    @classmethod
    def persist_set(cls, name, value, encoding=False, *, cfgfile=None):
        """ python里默认是改不了系统变量的，需要使用一些特殊手段
        https://stackoverflow.com/questions/17657686/is-it-possible-to-set-an-environment-variable-from-python-permanently/17657905

        :param cfgfile: 在linux系统时，可以使用该参数
            默认是把环境变量写入 ~/.bashrc，可以考虑写到
            TODO 有这个设想，但很不好实现，不是很关键的功能，所以还未开发

        """
        # 写入环境变量这里是有点小麻烦的，要考虑unix和windows不同平台，以及怎么持久化存储的问题，这里直接调用一个三方库来解决
        from envariable import setenv

        value = cls.set(name, value, encoding)
        if value[0] == value[-1] == '"':
            value = '\\' + value + '\\'
        setenv(name, value)

        return value

    @classmethod
    def unset(cls, name):
        """ 删除环境变量 """
        from envariable import unsetenv
        unsetenv(name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/11/05

"""
微信 消息框中不同消息结构的解析器 工具

todo 这个文件很多值可以写成枚举类型更规范？
"""

import re
import json

# 定义全局字典
msg_parsers = {}


# 定义装饰器
def register_tool(key):
    def decorator(func):
        # 将函数添加到全局字典中，键为传入的key参数
        msg_parsers[key] = func
        return func

    return decorator


def fmt_text(text):
    return text.replace('\n', ' ')


def parse_time_string(time_str):
    """
    解析不同格式的时间字符串为标准的 Python datetime 类型。

    支持以下格式：
    1. 完整的日期和时间，如 "2024年10月31日 22:03"，解析为对应的 datetime 对象。
    2. 仅包含时间的字符串，如 "19:18"，会被解析为当天的时间点，日期部分设为当前日期。
    3. 特殊字符串 "以下为新消息"，解析为当前时间的 datetime 对象。

    示例输入：
    - "2024年10月31日 22:03"
    - "2024年11月1日 20:09"
    - "2024年11月1日 21:08"
    - "19:18"
    - "20:16"
    - "以下为新消息"
    """
    from datetime import datetime

    # 定义当前日期
    now = datetime.now()

    # 定义不同的正则表达式模式
    full_date_pattern = r"(\d{4})年(\d{1,2})月(\d{1,2})日\s+(\d{1,2}):(\d{2})"
    partial_time_pattern = r"^(\d{1,2}):(\d{2})$"
    new_message_pattern = r"^以下为新消息$"

    # 解析完整的日期时间
    match = re.match(full_date_pattern, time_str)
    if match:
        year, month, day, hour, minute = map(int, match.groups())
        return datetime(year, month, day, hour, minute)

    # 解析只有时间（小时:分钟），默认当天的日期
    match = re.match(partial_time_pattern, time_str)
    if match:
        hour, minute = map(int, match.groups())
        return datetime(now.year, now.month, now.day, hour, minute)

    # 解析 "以下为新消息" 为当前时间
    if re.match(new_message_pattern, time_str):
        return now

    # 如果未匹配到任何模式，返回 None 表示无法解析
    return None


@register_tool("1b2p3p3b3p")
def 系统_查看更多消息(node):
    node.msg_type = 'system'
    node.content_type = 'button_more'
    node.render_text = f'⚙️: 点击此处可 {node.text}'


@register_tool("1l2t")
def 系统_时间标签(node):
    node.msg_type = 'system'
    node.content_type = 'time'
    node.time = parse_time_string(node.text)
    node.render_text = f'⚙️: {node.text}'


# 刚刚发送的特殊时间标签
@register_tool("1l2p3p3p3t3p3p")
def 系统_时间标签2(node):
    node.msg_type = 'system'
    node.content_type = 'time'
    node.time = parse_time_string(node.text)
    node.render_text = f'⚙️: {node.text}'


@register_tool("1l2p3p3t3p")
def 系统_撤回消息(node):
    node.msg_type = 'system'

    if node.text.endswith('撤回了一条消息'):
        node.content_type = 'recall'
        if node.text == '你撤回了一条消息':
            node.user = node.user2 = '你'
        else:
            node.user = re.search(r'"(.+)"', node.text).group(1)
            node.user2 = node.user
        node.render_text = f'⚙️: {node.text}'
    elif node.text.endswith('现在可以开始聊天了。'):
        node.content_type = 'add_friend'
        node.user = re.search(r'你已添加了(.+?)，现在可以开始聊天了。', node.text).group(1)
        node.user2 = node.user
        node.render_text = f'⚙️: {node.text}'
    else:
        raise ValueError


@register_tool("1l2p3p3p4p5p6p7t3b")
def 发送_文本(node):
    node.msg_type = 'send'
    node.content_type = 'text'
    node.user = node[0][2].text
    node.render_text = f'↑{node.user}: {fmt_text(node.text)}'


@register_tool("1l2p3p3p4p5p6p7p7p8p8b3b")
def 发送_图片(node):
    node.msg_type = 'send'
    node.content_type = 'image'
    node.user = node[0][2].text


@register_tool("1l2p3p3p4p5p6p7p7p8p9t9p9b8b3b")
def 发送_视频(node):
    node.msg_type = 'send'
    node.content_type = 'video'
    node.user = node[0][2].text


@register_tool("1l2p3p3p4p5p6p7p7p8p9p10p11t6p7p8p9p10p11t7b3b")
def 发送_文本_引用文本(node):
    node.msg_type = 'send'
    node.content_type = 'text'
    node.user = node[0][2].text
    quoted_node = node[0][1][0][0]  # 引用的层级结构
    node.text = quoted_node[0][1][0][0][0][0].text
    node.cite_text = quoted_node[1][0][0][0][0][0].text


@register_tool("1l2p3p3p4p5p6p7p7p8p9p10p11t6p7p8p9p10p11t11t11p12b7b3b")
def 发送_文本_引用图片(node):
    node.msg_type = 'send'
    node.content_type = 'text'
    node.user = node[0][2].text
    quoted_node = node[0][1][0][0]
    node.text = quoted_node[0][1][0][0][0][0].text
    node.cite_text = f'{quoted_node[1][0][0][0][0][0].text} : [图片]'


# 引用文件
@register_tool("1l2p3p3p4p5p6p7p7p8p9p10p11t6p7p8p9p10p11p12t11p12b7b3b")
def 发送_文本_引用文件(node):
    node.msg_type = 'send'
    node.content_type = 'text'
    node.user = node[0][2].text
    quoted_node = node[0][1][0][0]
    node.text = quoted_node[0][1][0][0][0][0].text
    node.cite_text = f'[文件] {quoted_node[1][0][0][0][0][0][0].text}'


# 发送文件
@register_tool("1l2p3p3p4p5p6p7p8p9p10t10p11t11t9p10p10p8p9t7b7b3b")
def 发送_文本_引用文件(node):
    node.msg_type = 'send'
    node.content_type = 'file'
    node.user = node[0][2].text
    file_node = node[0][1][0][0][0][0]
    desc = {
        'name': file_node[0][0][0].text,
        'size': file_node[0][0][1][0].text,
        'platform': file_node[1][0].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


# 发送链接
@register_tool("1l2p3p3p4p5p6p7p8t8p9t9p9b8p9b9t7b3b")
def 发送_链接(node):
    node.msg_type = 'send'
    node.content_type = 'link'
    node.user = node[0][2].text
    link_node = node[0][1][0][0]
    desc = {
        'title': link_node[0][0][0].text,
        'head': link_node[0][0][1][0].text,
        'author': link_node[0][0][2][1].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


# 转发消息
@register_tool("1l2p3p3p4p5p6p7p8t8p9t9t9t8p9p9t7b3b")
def 发送_转发消息(node):
    node.msg_type = 'send'
    node.content_type = 'messages'
    node.user = node[0][2].text


@register_tool("1l2p3b3p4p5p6p7t3p")
def 接收_文本(node):
    node.msg_type = 'receive'
    node.content_type = 'text'
    node.user = node[0][0].text
    node.render_text = f'↓{node.user}: {fmt_text(node.text)}'


@register_tool("1l2p3b3p4p5p6p7p7b3p")
def 接收_图片(node):
    node.msg_type = 'receive'
    node.content_type = 'image'
    node.user = node[0][0].text
    node.render_text = f'↓{node.user}: [图片]'


@register_tool("1l2p3b3p4p5p6p7p8p9p10t10p11t11t9p10p10p8p9t7b7b3p")
def 接收_文件(node):
    node.msg_type = 'receive'
    node.content_type = 'file'
    node.user = node[0][0].text
    node.render_text = f'↓{node.user}: {fmt_text(node.text)}'


@register_tool("1l2p3b3p4p5p6p7p8b8p8t7b3p")
def 接收_语音(node):
    node.msg_type = 'receive'
    node.content_type = 'voice'
    node.user = node[0][0].text


@register_tool("1l2p3b3p4p5p6p7p8t8p9t9p9b8p9b9t7b3p")
def 接收_链接(node):
    node.msg_type = 'receive'
    node.content_type = 'link'
    node.user = node[0][0].text
    link_node = node[0][1][0][0][0][0]  # 链接内容的层级结构
    desc = {
        'title': link_node[0].text,
        'head': link_node[1][0].text,  # 文章开头部分
        'author': link_node[2][1].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7t3p")
def 群接收_文本(node):
    node.msg_type = 'receive'
    node.content_type = 'text'
    node.user = node[0][0].text
    node.user2 = node[0][1][0][0].text


@register_tool("1l2p3b3p4p5t4p5p6p7p7b3p")
def 群接收_图片(node):
    node.msg_type = 'receive'
    node.content_type = 'image'
    node.user = node[0][0].text
    node.user2 = node[0][1][0][0].text


@register_tool("1l2p3b3p4p5t4p5p6p7p8p9p10t10p11t11t9p10p10p7b7b3p")
def 群接收_文件(node):
    node.msg_type = 'receive'
    node.content_type = 'file'
    node.user = node[0][0].text
    file_node = node[0][1]  # 文件信息的节点
    node.user2 = file_node[0][0].text
    file_info_node = file_node[1][0][0][0][0]
    desc = {
        'name': file_info_node[0][0].text,
        'size': file_info_node[0][1][0].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7p8p9p10t10p11t11t9p10p10p8p9t7b7b3p")
def 群接收_文件2(node):
    node.msg_type = 'receive'
    node.content_type = 'file'
    node.user = node[0][0].text
    node.user2 = node[0][1][0][0].text
    file_info_node = node[0][1][1][0][0][0]
    desc = {
        'name': file_info_node[0][0][0].text,
        'size': file_info_node[0][0][1][0].text,
        'platform': file_info_node[1][0].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7b3p")
def 群接收_动画表情(node):
    node.msg_type = 'receive'
    node.content_type = 'emoji'
    node.user = node[0][0].text
    node.user2 = node[0][1][0][0].text


@register_tool("1l2p3b3p4p5t4p5p6p7p8t8p9t9p9b8p9b9t7b3p")
def 群接收_链接_有公众号作者名(node):
    node.msg_type = 'receive'
    node.content_type = 'link'
    node.user = node[0][0].text
    link_node = node[0][1]
    node.user2 = link_node[0][0].text
    link_info_node = link_node[1][0][0][0]
    desc = {
        'title': link_info_node[0].text,
        'head': link_info_node[1][0].text,  # 文章开头的一小部分
        'author': link_info_node[2][0].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7p8t8p9t9p9b7b3p")
def 群接收_链接_无公众号作者名(node):
    node.msg_type = 'receive'
    node.content_type = 'link'
    node.user = node[0][0].text
    link_node = node[0][1]
    node.user2 = link_node[0][0].text
    link_info_node = link_node[1][0][0][0]
    desc = {
        'title': link_info_node[0].text,
        'head': link_info_node[1][0].text  # 文章开头的一小部分
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7p8b8p9p8p9p10p10t8p9p7b3p")
def 群接收_视频(node):
    node.msg_type = 'receive'
    node.content_type = 'video'
    node.user = node[0][0].text
    video_node = node[0][1]
    node.user2 = video_node[0][0].text
    video_info_node = video_node[1][0][0][0]
    desc = {
        'author': video_info_node[2][0][1].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7p8t8p9t8t7b3p")
def 群接收_视频2(node):
    node.msg_type = 'receive'
    node.content_type = 'video'
    node.user = node[0][0].text
    video_node = node[0][1]
    node.user2 = video_node[0][0].text
    video_info_node = video_node[1][0][0][0]
    desc = {
        'title': video_info_node[0].text,
        'head': video_info_node[1][0].text,
        'duration': video_info_node[2].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3b3p4p5t4p5p6p7p8p9b9t8t8b8p9t9t7b3p")
def 群接收_小程序(node):
    node.msg_type = 'receive'
    node.content_type = 'applet'
    node.user = node[0][0].text
    applet_node = node[0][1]
    node.user2 = applet_node[0][0].text
    applet_info_node = applet_node[1][0][0][0]
    desc = {
        'name': applet_info_node[0][1].text,
        'title': applet_info_node[1].text,
        'footnote': applet_info_node[3][1].text
    }
    node.text = json.dumps(desc, ensure_ascii=False)


@register_tool("1l2p3p3p4p5l6p7p7t7p3p")
def 群接收_拍一拍(node):
    node.msg_type = 'receive'
    node.content_type = 'shake'
    node.text = node[0][1][0][0].text
    node.user = re.search(r'"(.+?)" 拍了拍 "', node.text).group(1)


@register_tool("1l2p3b3p4p5t4p5p6p7p8t4p5p6p7p8p9t5b3p")
def 群接收_引用消息(node):
    node.msg_type = 'receive'
    node.content_type = 'text'
    node.user = node[0][0].text
    quoted_node = node[0][1]
    node.user2 = quoted_node[0][0].text
    node.text = quoted_node[1][0][0][0][0].text
    node.cite_text = quoted_node[2][0][0][0][0][0].text


@register_tool("1l2p3b3p4p5t4p5p6p7p8t4p5p6p7p8p9p10t9p10b5b3p")
def 群接收_引用文件(node):
    node.msg_type = 'receive'
    node.content_type = 'text'
    node.user = node[0][0].text
    quoted_node = node[0][1]
    node.user2 = quoted_node[0][0].text
    node.text = quoted_node[1][0][0][0][0].text
    file_node = quoted_node[2][0][0][0][0][0][0]
    node.cite_text = f'[文件] {file_node.text}'



import sys
from .chat import Chat, AKPool
from arts.cooltypes import moduledb


mdb = moduledb.DB('https://pypi.org/project/openai2', depth=3)['chat_in_cmd']
apikeys: moduledb.File = mdb['apikeys']['data_1']
apikeys.setdefault('list', [])
record: moduledb.File = mdb['records']['chat_1']
record.setdefault('messages', [])


def chat_in_cmd(apikeys:list, newchat=False, model='gpt-4-1106-preview', msg_max_count=30):
    print(f"\n\033[0m你已进入命令行聊天模式, 当前使用'{model}'模型, 请确保你的apikey支持该模型.", end='')
    gpt = Chat(api_key=AKPool(apikeys), model=model, msg_max_count=msg_max_count)
    if not newchat:
        gpt.add_dialogs(*record['messages'])
    while True:
        user = input('\n\n\033[32;1m:')
        print('\033[0m')
        for x in gpt.stream_request(user):
            print(x, end='', flush=True)
        record['messages'] = gpt.fetch_messages()


命令提示 = '''指令集:
openai2 add_apikey <apikey> | 添加1个apikey              | 如需添加多个, 可执行多次
openai2 read_apikey         | 查看所有apikey             |
openai2 clear_apikey        | 清除所有apikey             |
openai2 chat                | 继续上次的对话              |
openai2 newchat             | 清空对话记录, 然后开始新对话 |
'''


def ParseCmd():
    kws = sys.argv[1:]
    if kws:
        kw = kws[0].lower()
        
        # 添加apikey
        if kw == 'add_apikey' and len(kws) > 1:
            apikeys['list'].append( kws[1] )
            apikeys.save()
            print('添加 apikey 成功.')

        # 继续上次的对话
        elif kw == 'chat':
            chat_in_cmd(apikeys['list'], newchat=False)
        
        # 清空对话记录, 然后开始新对话
        elif kw == 'newchat':
            chat_in_cmd(apikeys['list'], newchat=True)
        
        # 查看所有apikey
        elif kw == 'read_apikey':
            for x in apikeys['list']:
                print(x)
        
        # 清除所有apikey
        elif kw == 'clear_apikey':
            apikeys['list'] = []
            print('已清除所有 apikey .')

        else:
            print(命令提示)
    else:
        print(命令提示)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/03 21:17

import os
import re
import shutil
import struct


def struct_unpack(f, fmt):
    r""" 类似np.fromfile的功能，读取并解析二进制数据

    :param f:
        如果带有read方法，则用read方法读取指定字节数
        如果bytes对象则直接处理
    :param fmt: 格式
        默认按小端解析(2, 1, 0, 0) -> 258，如果需要大端，可以加前缀'>'
        字节：c=char, b=signed char, B=unsigned char, ?=bool
        2字节整数：h=short, H=unsigned short（后文同理，大写只是变成unsigned模式，不在累述）
        4字节整数：i, I, l, L
        8字节整数：q, Q
        浮点数：e=2字节，f=4字节，d=8字节

    >>> b = struct.pack('B', 127)
    >>> b
    b'\x7f'
    >>> struct_unpack(b, 'c')
    b'\x7f'
    >>> struct_unpack(b, 'B')
    127

    >>> b = struct.pack('I', 258)
    >>> b
    b'\x02\x01\x00\x00'
    >>> struct_unpack(b, 'I')  # 默认都是按小端打包、解析
    258
    >>> struct_unpack(b, '>I') # 错误示范，按大端解析的值
    33619968
    >>> struct_unpack(b, 'H'*2)  # 解析两个值，fmt*2即可
    (258, 0)

    >>> f = io.BytesIO(b'\x02\x01\x03\x04')
    >>> struct_unpack(f, 'B'*3)  # 取前3个值，等价于np.fromfile(f, dtype='uint8', count=3)
    (2, 1, 3)
    >>> struct_unpack(f, 'B')  # 取出第4个值
    4
    """
    # 1 取数据
    size_ = struct.calcsize(fmt)
    if hasattr(f, 'read'):
        data = f.read(size_)
        if len(data) < size_:
            raise ValueError(f'剩余数据长度 {len(data)} 小于 fmt 需要的长度 {size_}')
    else:  # 对于bytes等矩阵，可以多输入，但是只解析前面一部分
        data = f[:size_]

    # 2 解析
    res = struct.unpack(fmt, data)
    if len(res) == 1:  # 解析结果恰好只有一个的时候，返回值本身
        return res[0]
    else:
        return res


def recreate_folders(*dsts):
    """重建一个空目录"""
    for dst in dsts:
        try:
            # 删除一个目录（含内容），设置ignore_errors可以忽略目录不存在时的错误
            shutil.rmtree(dst, ignore_errors=True)
            os.makedirs(dst)  # 重新新建一个目录，注意可能存在层级关系，所以要用makedirs
        except TypeError:
            pass


def checkpathfile(name):
    r"""判断环境变量path下是否有name这个文件，有则返回绝对路径，无则返回None
    常用的有：BCompare.exe、Chrome.exe、mogrify.exe、xelatex.exe

    >> checkpathfile('xelatex.exe')
    'C:\\CTEX\\MiKTeX\\miktex\\bin\\xelatex.exe'
    >> checkpathfile('abcd.exe')
    """
    for path in os.getenv('path').split(';'):
        fn = os.path.join(path, name)
        if os.path.exists(fn):
            return fn
    return None


def filename_tail(fn, tail):
    """在文件名末尾和扩展名前面加上一个tail"""
    names = os.path.splitext(fn)
    return names[0] + tail + names[1]


def hasext(f, *exts):
    """判断文件f是否是exts扩展名中的一种，如果不是返回False，否则返回对应的值

    所有文件名统一按照小写处理
    """
    ext = os.path.splitext(f)[1].lower()
    exts = tuple(map(lambda x: x.lower(), exts))
    if ext in exts:
        return ext
    else:
        return False


def isdir(fn):
    """判断输入的是不是合法的路径格式，且存在确实是一个文件夹"""
    try:
        return os.path.isdir(fn)
    except ValueError:  # 出现文件名过长的问题
        return False
    except TypeError:  # 输入不是字符串类型
        return False


__mygetfiles = """
py有os.walk可以递归遍历得到一个目录下的所有文件
但是“我们”常常要过滤掉备份文件（171020-153959），Old、temp目、.git等目录
特别是windows还有一个很坑爹的$RECYCLE.BIN目录。
所以在os.walk的基础上，再做了封装得到myoswalk。

然后在myoswalk基础上，实现mygetfiles。
"""


def gen_file_filter(s):
    """生成一个文件名过滤函数"""
    if s[0] == '.':
        return lambda x: x.endswith(s)
    else:
        s = s.replace('？', r'[\u4e00-\u9fa5]')  # 中文问号可以匹配任意中文字符
        return lambda x: re.search(s, x)


def getfiles(root, filter_rule=None):
    r""" 对os.walk进一步封装，返回所有匹配的文件

    可以这样遍历一个目录下的所有文件：
    for f in getfiles(r'C:\pycode\code4101py', r'.py'):
        print(f)
    筛选规则除了“.+后缀”，还可以写正则匹配
    """
    if isinstance(filter_rule, str):
        filter_rule = gen_file_filter(filter_rule)

    for root, _, files in os.walk(root, filter_rule):
        for f in files:
            if filter_rule and not filter_rule(f):
                continue
            yield root + '\\' + f


def tex_content_filefilter(f):
    """只获取正文类tex文件"""
    if f.endswith('.tex') and 'Conf' not in f and 'settings' not in f:
        return True
    else:
        return False


def tex_conf_filefilter(f):
    """只获取配置类tex文件"""
    if f.endswith('.tex') and ('Conf' in f or 'settings' in f):
        return True
    else:
        return False


def change_ext(filename, ext):
    """更改文件名后缀
    返回第1个参数是新的文件名，第2个参数是这个文件是否存在

    输入的fileName可以没有扩展名，如'A/B/C/a'，仍然可以找对应的扩展名为ext的文件
    输入的ext不要含有'.'，例如正确格式是输入'tex'、'txt'
    """
    name = os.path.splitext(filename)[0]  # 'A/B/C/a.txt' --> 'A/B/C/a'
    newname = name + '.' + ext
    return newname, os.path.exists(newname)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, "..")))

import argparse

import paddle
from paddle.jit import to_static

from pyxlpr.ppocr.modeling.architectures import build_model
from pyxlpr.ppocr.postprocess import build_post_process
from pyxlpr.ppocr.utils.save_load import load_model
from pyxlpr.ppocr.utils.logging import get_logger
from pyxlpr.ppocr.tools.program import load_config, merge_config, ArgsParser


def export_single_model(model, arch_config, save_path, logger):
    if arch_config["algorithm"] == "SRN":
        max_text_length = arch_config["Head"]["max_text_length"]
        other_shape = [
            paddle.static.InputSpec(
                shape=[None, 1, 64, 256], dtype="float32"), [
                    paddle.static.InputSpec(
                        shape=[None, 256, 1],
                        dtype="int64"), paddle.static.InputSpec(
                            shape=[None, max_text_length, 1], dtype="int64"),
                    paddle.static.InputSpec(
                        shape=[None, 8, max_text_length, max_text_length],
                        dtype="int64"), paddle.static.InputSpec(
                            shape=[None, 8, max_text_length, max_text_length],
                            dtype="int64")
                ]
        ]
        model = to_static(model, input_spec=other_shape)
    elif arch_config["algorithm"] == "SAR":
        other_shape = [
            paddle.static.InputSpec(
                shape=[None, 3, 48, 160], dtype="float32"),
        ]
        model = to_static(model, input_spec=other_shape)
    else:
        infer_shape = [3, -1, -1]
        if arch_config["model_type"] == "rec":
            infer_shape = [3, 32, -1]  # for rec model, H must be 32
            if "Transform" in arch_config and arch_config[
                    "Transform"] is not None and arch_config["Transform"][
                        "name"] == "TPS":
                logger.info(
                    "When there is tps in the network, variable length input is not supported, and the input size needs to be the same as during training"
                )
                infer_shape[-1] = 100
            if arch_config["algorithm"] == "NRTR":
                infer_shape = [1, 32, 100]
        elif arch_config["model_type"] == "table":
            infer_shape = [3, 488, 488]
        model = to_static(
            model,
            input_spec=[
                paddle.static.InputSpec(
                    shape=[None] + infer_shape, dtype="float32")
            ])

    paddle.jit.save(model, save_path)
    logger.info("inference model is saved to {}".format(save_path))
    return


def main(config, logger):
    # FLAGS = ArgsParser().parse_args()
    # config = load_config(FLAGS.config)
    # merge_config(FLAGS.opt)
    # logger = get_logger()
    # build post process

    post_process_class = build_post_process(config["PostProcess"],
                                            config["Global"])

    # build model
    # for rec algorithm
    if hasattr(post_process_class, "character"):
        char_num = len(getattr(post_process_class, "character"))
        if config["Architecture"]["algorithm"] in ["Distillation",
                                                   ]:  # distillation model
            for key in config["Architecture"]["Models"]:
                config["Architecture"]["Models"][key]["Head"][
                    "out_channels"] = char_num
                # just one final tensor needs to to exported for inference
                config["Architecture"]["Models"][key][
                    "return_all_feats"] = False
        else:  # base rec model
            config["Architecture"]["Head"]["out_channels"] = char_num
    model = build_model(config["Architecture"])
    load_model(config, model)
    model.eval()

    save_path = config["Global"]["save_inference_dir"]

    arch_config = config["Architecture"]

    if arch_config["algorithm"] in ["Distillation", ]:  # distillation model
        archs = list(arch_config["Models"].values())
        for idx, name in enumerate(model.model_name_list):
            sub_model_save_path = os.path.join(save_path, name, "inference")
            export_single_model(model.model_list[idx], archs[idx],
                                sub_model_save_path, logger)
    else:
        save_path = os.path.join(save_path, "inference")
        export_single_model(model, arch_config, save_path, logger)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/09/18 22:16

import os
import concurrent.futures
import math
import time
import sys

from pyxllib.prog.pupil import EmptyPoolExecutor, format_exception
from pyxllib.text.pupil import shorten

XLLOG_CONF_FILE = 'xllog.yaml'

____xllog = """
"""


def get_xllog(name='xllog', *, log_file=None):
    """ 获得pyxllib库的日志类

    :param log_file: 增加输出到一个日志文件，该功能仅在首次初始化时有效
        注意这个是'w'机制，会删除之前的日志文件
        # TODO 这样的功能设计问题挺大的，工程逻辑很莫名其妙，有空要把日志功能修缮下
        #   例如一个通用的初始化类，然后xllog只是一个特定的实例日志类

    TODO 增加输出到钉钉机器人、邮箱的Handler？
    """
    # import logging, coloredlogs
    import logging

    # 1 判断是否已存在，直接返回
    if ('pyxllib.' + name) in logging.root.manager.loggerDict:
        return logging.getLogger('pyxllib.' + name)

    # 2 初次构建
    if name == 'xllog':  # 附带运行时间信息
        if os.path.isfile(XLLOG_CONF_FILE):
            # 尝试在默认位置是否有自定义配置文件，读取配置文件来创建
            import logging.config
            from pyxllib.file.specialist import File
            data = File(XLLOG_CONF_FILE).read()
            if isinstance(data, dict):
                # 推荐使用yaml的字典结构，格式更简洁清晰
                logging.config.dictConfig(data)
            else:
                # 但是普通的conf配置文件也支持
                logging.config.fileConfig(XLLOG_CONF_FILE)
        else:
            # 3 否则生成一个非常简易版的xllog
            xllog = logging.getLogger('pyxllib.xllog')
            fmt = '%(asctime)s %(message)s'
            if log_file:
                file_handler = logging.FileHandler(f'{log_file}', 'a')
                file_handler.setLevel(logging.DEBUG)
                file_handler.setFormatter(logging.Formatter(fmt))
                xllog.addHandler(file_handler)
            # coloredlogs.install(level='DEBUG', logger=xllog, fmt=fmt)
    elif name == 'location':  # 附带代码所处位置信息
        loclog = logging.getLogger('pyxllib.location')
        # coloredlogs.install(level='DEBUG', logger=loclog, fmt='%(filename)s/%(lineno)d: %(message)s')
    return logging.getLogger('pyxllib.' + name)


class Iterate:
    """ 迭代器类，用来封装一些特定模式的for循环操作

    TODO 双循环，需要内部两两对比的迭代功能

    200920周日18:20，最初设计的时候，是提供run_pair、run_pair2的功能的
        不过后来想想，这个其实就是排列组合，在itertools里有combinations, permutations可以代替
        甚至有放回的组合也有combinations_with_replacement，我实在是不需要再这里写这些冗余的功能
        所以就移除了
    """

    def __init__(self, items):
        # 没有总长度倒也能接受，关键是可能要用start、end切片，所以还是先转成tuple更方便操作
        self.items = tuple(items)
        self.n_items = len(self.items)
        self.format_width = math.ceil(math.log10(self.n_items + 1))
        self.xllog = get_xllog()

    def _format_pinterval(self, pinterval=None):
        if isinstance(pinterval, str) and pinterval.endswith('%'):
            # 百分比的情况，重算出间隔元素数
            return int(round(self.n_items * float(pinterval[:-1]) / 100))
        else:  # 其他格式暂不解析，按原格式处理
            return pinterval

    def _step1_check_number(self, pinterval, func):
        if pinterval:
            sys.stdout.flush()  # 让逻辑在前的标准输出先print出来，但其实这句也不一定能让print及时输出的~~可能会被日志提前抢输出了
            self.xllog.info(f"使用 {func.__name__} 处理 {self.n_items} 个数据 {shorten(str(self.items), 30)}")

    def _step2_check_range(self, start, end):
        if start:
            self.xllog.info(f"使用start参数，只处理≥{start}的条目")
        else:
            start = 0
        if end:
            # 这里空格是为了对齐，别删
            self.xllog.info(f"使用 end 参数，只处理<{end}的条目")
        else:
            end = len(self.items)
        return start, end

    def _step3_executor(self, pinterval, max_workers):
        if max_workers == 1:
            # workers=1，实际上并不用多线程，用一个假的多线程类代替，能大大提速
            executor = EmptyPoolExecutor()
            # executor = concurrent.futures.ThreadPoolExecutor(max_workers)
        else:
            executor = concurrent.futures.ThreadPoolExecutor(max_workers)
            if pinterval:
                self.xllog.info(f'多线程执行，当前迭代所用线程数：{executor._max_workers}')
        return executor

    def _step4_iter(self, i, pinterval, executor):
        # 队列中没有新任务时，才放入新任务，这样能确保pinterval的输出能反应实时情况，而不是一下全部进入队列，把for循环跑完了
        while executor._work_queue.qsize(): pass
        if pinterval and (i or pinterval == 1) and i % pinterval == 0:
            message = f' {self.items[i]}' if pinterval == 1 else ''
            self.xllog.info(f'{i:{self.format_width}d}/{self.n_items}={i / self.n_items:6.2%}{message}')

    def _step5_finish(self, pinterval, interrupt, start_time):
        from humanfriendly import format_timespan
        end_time = time.time()
        span = end_time - start_time
        if span:
            speed = self.n_items / span
            msg = f'总用时：{format_timespan(span)}，速度：{speed:.2f}it/s'
        else:
            msg = f'总用时：{format_timespan(span)}'
        if not interrupt and pinterval:
            self.xllog.info(f'{self.n_items / self.n_items:6.2%} 完成迭代，{msg}')
            sys.stderr.flush()

    def run(self, func, start=0, end=None, pinterval=None, max_workers=1, interrupt=True):
        """
        :param func: 对每个item执行的功能
        :param start: 跳过<start的数据，只处理>=start编号以上
        :param end: 只处理 < end 的数据
        :param pinterval: 每隔多少条目输出进度日志，默认不输出进度日志（但是错误日志依然会输出）
            支持按百分比进度显示，例如每20%，pinterval='20%'，不过一些底层实现机制原因，会有些许误差
            TODO 支持按指定时间间隔显示？ 例如每15秒，pinterval='15s' 感觉这种功能太花哨了，没必要搞
        :param max_workers: 默认线程数，默认1，即串行
        :type max_workers: int, None
        :param interrupt: 出现错误时是否中断，默认True会终止程序，否则只会输出错误日志
        :return:
        """

        # 1 统一的参数处理部分
        pinterval = self._format_pinterval(pinterval)
        self._step1_check_number(pinterval, func)
        start, end = self._step2_check_range(start, end)
        error = False
        executor = self._step3_executor(pinterval, max_workers)

        # 2 封装的子处理部分
        def wrap_func(func, i):
            nonlocal error
            item = self.items[i]
            try:
                func(item)
            except Exception as e:
                error = e
                self.xllog.error(f'💔idx={i}运行出错：{item}\n{format_exception(e)}')

        # 3 执行迭代
        start_time = time.time()
        for i in range(start, end):
            self._step4_iter(i, pinterval, executor)
            executor.submit(wrap_func, func, i)
            if interrupt and error:
                raise error
        executor.shutdown()  # 必须等executor结束，error才是准确的
        self._step5_finish(pinterval, interrupt and error, start_time)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/05/10

from pyxllib.prog.pupil import check_install_package

check_install_package('tiktoken')

import re

import tiktoken

from pyxllib.prog.newbie import SingletonForEveryInitArgs


class TicToken(metaclass=SingletonForEveryInitArgs):
    def __init__(self, for_model='gpt-4'):
        """ gpt4，默认使用的就是cl100k_base """
        self.for_model = for_model
        self.enc = tiktoken.encoding_for_model(for_model)

    def encode(self, text):
        """ 将文本转换为token """
        return self.enc.encode(text)

    def count_tokens(self, text, max_length=20000, num_segments=4, *, delete_space=False):
        """ 统计文本的token数量，对于超过max_length的文本，采用估算方法 """
        if delete_space:
            text = re.sub(r'\s+', '', text)

        if len(text) <= max_length:
            return len(self.encode(text))

        # 对长文本进行采样估计
        total_length = len(text)
        sample_size = max_length // num_segments
        interval = (total_length - sample_size) // (num_segments - 1)
        samples = [text[i * interval: i * interval + sample_size] for i in range(num_segments)]
        sample_tokens = sum(len(self.encode(sample)) for sample in samples)
        estimated_tokens = sample_tokens * total_length // max_length
        return estimated_tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/04/01

from collections import Counter

from pyxllib.prog.pupil import check_install_package

check_install_package('jmespath')

import jmespath


class JsonTool:
    @classmethod
    def find_paths_by_condition(cls, obj, condition, current_path=''):
        """
        根据提供的条件递归搜索JSON对象，查找满足条件的路径。
        """
        if isinstance(obj, dict):
            for k, v in obj.items():
                new_path = f"{current_path}.{k}" if current_path else k
                if condition(k, v):
                    yield new_path
                else:
                    if isinstance(v, (dict, list)):
                        yield from JsonTool.find_paths_by_condition(v, condition, new_path)
        elif isinstance(obj, list):
            for index, item in enumerate(obj):
                new_path = f"{current_path}[{index}]"
                if isinstance(item, (dict, list)):
                    yield from JsonTool.find_paths_by_condition(item, condition, new_path)

    @classmethod
    def find_paths_with_key(cls, obj, key, current_path=''):
        """
        查找包含指定键名的路径。
        """
        # 定义一个简单的条件函数，仅基于键名
        condition = lambda k, v: k == key
        return JsonTool.find_paths_by_condition(obj, condition, current_path)


class JsonlTool:
    @classmethod
    def check_key_positions(cls, json_list, key):
        """
        在JSON对象列表中统计每种包含指定键名的路径出现的次数。

        :param json_list: 包含多个JSON对象的列表。
        :param key: 要查找的键名。
        :return: 一个Counter对象，统计每种路径出现的次数。
        """
        paths_counter = Counter()
        for json_obj in json_list:
            # 使用JsonTool查找当前对象中包含指定键的路径
            paths = list(JsonTool.find_paths_with_key(json_obj, key))
            paths_counter.update(paths)

        return paths_counter

    @classmethod
    def check_path_values(cls, json_list, jmespath_expression, value_extractor=lambda x: x):
        """
        根据提供的jmespath表达式，在JSON对象列表中直接统计表达式取值的次数。
        支持自定义函数来获取要统计的键。

        :param json_list: 包含多个JSON对象的列表。
        :param jmespath_expression: 要应用的jmespath表达式。
        :param value_extractor: 一个函数，用于从jmespath.search的结果中提取要统计的值。
        :return: 一个Counter对象，统计提取值出现的次数。
        """
        values_counter = Counter()
        for json_obj in json_list:

            # 使用jmespath搜索当前对象
            result = jmespath.search(jmespath_expression, json_obj)
            # 使用自定义的value_extractor函数提取要统计的值
            extracted_values = value_extractor(result)
            if not isinstance(extracted_values, list):
                extracted_values = [extracted_values]

            for value in extracted_values:
                values_counter[value] += 1

        values_counter = Counter(dict(values_counter.most_common()))
        return values_counter



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .iaa_augment import IaaAugment
from .make_border_map import MakeBorderMap
from .make_shrink_map import MakeShrinkMap
from .random_crop_data import EastRandomCropData, RandomCropImgMask
from .make_pse_gt import MakePseGt

from .rec_img_aug import RecAug, RecResizeImg, ClsResizeImg, SRNRecResizeImg, NRTRRecResizeImg, SARRecResizeImg
from .randaugment import RandAugment
from .copy_paste import CopyPaste
from .ColorJitter import ColorJitter
from .operators import *  # 操作器的定义方式，可以参考这里的DecodeImage实现
from .label_ops import *

from .east_process import *
from .sast_process import *
from .pg_process import *
from .gen_table_mask import *


def transform(data, ops=None):
    """ transform """
    if ops is None:
        ops = []
    for op in ops:
        data = op(data)
        if data is None:
            return None
    return data


def create_operators(op_param_list, global_config=None):
    """
    create operators based on the config

    Args:
        params(list): a dict list, used to create some operators
    """
    assert isinstance(op_param_list, list), ('operator config should be a list')
    ops = []
    for operator in op_param_list:
        assert isinstance(operator,
                          dict) and len(operator) == 1, "yaml format error"
        op_name = list(operator)[0]
        param = {} if operator[op_name] is None else operator[op_name]
        if global_config is not None:
            param.update(global_config)
        op = eval(op_name)(**param)
        ops.append(op)
    return ops



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/04/12

import re
from urllib.parse import unquote
import tempfile
import time
from urllib.parse import urlparse

from deprecated import deprecated
from loguru import logger
import DrissionPage
from DrissionPage import ChromiumPage, Chromium
from DrissionPage._pages.chromium_base import ChromiumBase
from DrissionPage._pages.chromium_tab import ChromiumTab
import DrissionPage.errors

from pyxllib.prog.pupil import inject_members
from pyxllib.text.pupil import strfind
from pyxllib.file.specialist import GetEtag


@deprecated(reason='get_dp_page逻辑不太对，请换用get_dp_tab')
def get_dp_page(dp_page=None) -> 'XlPage':
    """

    :param dp_page:
        默认None, 返回默认的page，一般就是当前页面
        True, 新建一个page
        str, 新建一个对应url的page
        func(tab), 通过规则筛选tab，返回符合条件的第1个tab，否则新建一个tab
    """

    if isinstance(dp_page, Chromium):
        return dp_page
    elif isinstance(dp_page, ChromiumTab):
        return dp_page.page
    elif callable(dp_page):
        page0 = Chromium()
        for tab in page0.get_tabs():
            if dp_page(tab):
                return tab.page
        return page0.new_tab().page
    elif dp_page is True:
        return Chromium().new_tab().page
    elif isinstance(dp_page, str):
        return Chromium().new_tab(dp_page).page
    else:
        return Chromium()


def get_dp_tab(dp_tab=None) -> 'XlTab':
    """ 智能获取一个标签页tab

    :param dp_tab:
        默认None, 返回默认的tab，一般就是当前页面
        True, 新建一个tab
        str, 新建一个对应url的tab，如果发现已存在的tab中已经有该url，则直接复用
        func(tab), 通过规则筛选tab，返回符合条件的第1个tab，否则新建一个tab
    """

    if isinstance(dp_tab, Chromium):
        return dp_tab.latest_tab
    elif isinstance(dp_tab, ChromiumTab):
        return dp_tab
    elif callable(dp_tab):
        page0 = Chromium()
        for tab in page0.get_tabs():
            if dp_tab(tab):
                return tab
        return page0.new_tab()
    elif dp_tab is True:
        return Chromium().new_tab()
    elif isinstance(dp_tab, str):
        def dp_page2(tab):  # 默认开启页面复用
            if tab.url == dp_tab:
                return tab

        tab = get_dp_tab(dp_page2)
        if tab.url == 'about:blank':
            tab.get(dp_tab)
        return tab
    else:
        return Chromium().latest_tab


def get_latest_not_dev_tab(browser=None):
    """ 开发工具本身也会算一个tab，这个函数返回最新的一个不是开发工具的tab """
    if browser is None:
        browser = Chromium()
    tabs = browser.get_tabs()
    for tab in tabs:
        if strfind(tab.url, ['devtools://', 'chrome-extension://']) != -1:
            continue
        return tab


class XlChromiumBase(ChromiumBase):
    def get2(self, url, show_errmsg=False, retry=None, interval=None):
        """
        240418周四21:57，DrissionPage-4.0.4.21 官方自带page.get，有时候会有bug，不会实际刷新url，这里加个代码进行fix
        """
        old_url = self.url
        if old_url == url:
            # page.refresh()
            return

        ele = self.active_ele
        self.get(url, show_errmsg=show_errmsg, retry=retry, interval=interval)
        try:  # 如果新页面获取成功，理论上旧的ele会失效
            ele
            self.refresh()  # 如果不报错，这里网站要强制更新
            return self
        except DrissionPage.errors.ElementLostError:
            return self

    def get_download_files(self: Chromium):
        """ 获取下载列表

         (241205周四21:02，这个功能原本是用来做页面文件下载的，
        但后来知道dp有更简洁的解决方案后，其实原本功能意义已不大，
        只是作为一个结构化解析下载页面的功能，可以保留参考)

        :param search_name: 搜索文件名，输入该参数时，只会从上往下找到第一个匹配的文件
            否则返回一个list结构，存储下载清单里的文件
        :return:

        todo 默认应该显示是不全的，有时间可以考虑往下滑动继续检索的功能
        """
        files = []
        page2 = Chromium().new_tab('chrome://downloads/')
        items = page2('tag:downloads-manager', timeout=1).sr('#mainContainer')('#downloadsList').eles(
            'tag:downloads-item')
        for item in items:
            loc = unquote(item.sr('tag:img').attr('src').replace('+', ' '))
            file = re.search(r'path=(.+?)(&scale=(\d+(\.\d+)?)x)?$', loc).group(1)

            files.append({
                'file': file,
                'url': unquote(item.sr('#url').attr('href'))
            })

        page2.close()

        return files

    def wait_page_not_change(self, interval=3):
        """ 等待直到页面内容不再变化

        :param interval: 时间间隔，需要判断当前内容和interval秒后的内容，看内容是否欧发生改变
        """
        last_html, last_etag = None, None
        while True:
            html = self.html
            etag = GetEtag.from_text(html)
            if etag == last_etag:
                break

            last_html, last_etag = html, etag
            time.sleep(interval)
        return last_html

    def action_type(self, ele, text, clear=True):
        """ 基于action实现的重写入，常用于日期相关操作
        因为很多日期类组件，直接使用ele.input是不生效的，哪怕看似显示了文本，但其实并没有触发js改动，需要用动作链来实现
        """
        from DrissionPage.common import Keys
        if clear:
            self.actions.click(ele).key_down(Keys.CTRL).type('a').key_up(Keys.CTRL).type(text)
        else:
            self.actions.click(ele).type(text)


inject_members(XlChromiumBase, ChromiumBase)


class XlPage(XlChromiumBase, ChromiumTab, Chromium):
    """ 只作为一个类型标记，无实质功能。在猴子补丁背景下，让ide能正确跳转函数定义。 """
    pass


class XlTab(XlChromiumBase, ChromiumTab, Chromium):
    pass


def wait_page_not_change(page, interval=3):
    page.wait_page_not_change(interval)


class DpWebBase:
    """ 基于dp开发的爬虫工具的一个基础类 """

    def __init__(self, url=None, *, base_url=None):
        self.browser = Chromium()
        self.browser.set.download_path(tempfile.gettempdir())

        parsed_url = urlparse(url)
        root_url = f"{parsed_url.scheme}://{parsed_url.netloc}"  # 构建基础 URL
        self.tab: XlTab = self.browser.new_tab(url)
        self.base_url = base_url or root_url  # 使用基础 URL 作为 base_url，后续同域名网站去重用

    def close_if_exceeds_min_tabs(self, min_tabs_to_keep=1):
        """ 检查同网站的tab数量，如果超过最小保留数量则关闭当前页面 """
        try:  # 靠Py结束触发的可能报错：ImportError: sys.meta_path is None, Python is likely shutting down
            if self.tab and self.base_url and len(self.browser.get_tabs(url=self.base_url)) > min_tabs_to_keep:
                self.tab.close()
        except Exception as e:
            pass

    # 实测效果不稳定，感觉还不如手动触发吧~
    # def __del__(self):
    #     """ 我习惯每次新任务建立新的tab处理，并在结束后自动检查同网页打开的标签是否不唯一则删掉 """
    #     self.close_if_exceeds_min_tabs()


def close_duplicate_tabs(browser=None):
    """ 关闭浏览器重复标签页

    遍历所有标签页（从前往后，dp的get_tabs拿到的tab就是从新到旧的），对每个域名仅保留第一个出现的标签页，其余同域名标签页关闭；
    如果最后还剩多个标签页，则把'chrome://newtab/'也关掉。
    """
    # 1 初始化
    if browser is None:
        browser = Chromium()

    # 250115周三21:12 这步不稳定，会报错，不知道为啥。导致dp最后经常没有清理tabs
    # 250204周二09:21，好像是浏览器重启更新到最新版本就行了~ 这里也要加个try
    try:
        all_tabs = browser.get_tabs()
    except TimeoutError:
        logger.warning(
            'browser.get_tabs()运行报错，请清查浏览器是否已更新但没有重启。本次将browser.quit()退出整个浏览器。')
        # 你不让我关tabs是吧，那我就把整个浏览器关了
        browser.quit()
        return

    seen_domains = set()

    # 2 第一次遍历：保留首个出现的域名，其余重复则关闭
    for t in all_tabs:
        parsed_url = urlparse(t.url)
        domain = parsed_url.netloc  # netloc 通常可拿到域名部分
        # logger.info(f'{t.url}, {domain}')

        if domain in seen_domains:
            t.close()
        else:
            seen_domains.add(domain)

    # 3 第二次遍历：如果剩余标签页 > 1，则关掉chrome://newtab/
    remaining_tabs = browser.get_tabs()
    if len(remaining_tabs) > 1:
        for t in remaining_tabs:
            if t.url.startswith('chrome://newtab'):
                t.close()


def dp_check_quit():
    """ 检查当前页面是否只剩空标签页，则浏览器可以自动退出 """
    browser = Chromium()
    try:
        tabs = browser.get_tabs()
    except TimeoutError:
        logger.warning('browser.get_tabs()运行报错，浏览器可能已更新但没有重启。将退出浏览器。')
        browser.quit()
        return

    # 检查是否只剩下空标签页
    if len(tabs) == 1 and tabs[0].url.startswith('chrome://newtab'):
        # 如果只剩下一个空标签页，则退出浏览器
        browser.quit()
    elif len(tabs) == 0:
        # 如果没有标签页，也退出浏览器
        browser.quit()
